#  @author Vincenzo Eduardo Padulano
#  @author Enric Tejedor
#  @date 2021-02

################################################################################
# Copyright (C) 1995-2022, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################
from __future__ import annotations

from typing import Dict, Union

import ROOT


class Operation:
    """An operation attached to a distributed RDataFrame graph node."""

    def __init__(self, name: str, *args, **kwargs):
        self.name = name
        self.args = list(args)
        self.kwargs = kwargs


class Action(Operation):
    """An action attached to a distributed RDataFrame graph node."""
    pass


class Histo(Action):
    """
    Any type of Histo*D action.

    This distinct class is needed to check that the user passes a model for the
    histogram action. Merging histograms coming from different distributed tasks
    is not possible unless the model was previously specified, since the binning
    of the different histograms would be incompatible.
    """

    def __init__(self, name: str, *args, **kwargs):
        super().__init__(name, *args, **kwargs)

        # The histogram model can be passed as the keyword argument 'model'. All
        # Histo*D specializations have the same name for this argument. If it is
        # present, we know the execution can proceed safely.
        if "model" not in self.kwargs:
            # If the keyword argument was not passed, we need to check the first
            # positional argument. In all Histo*D overload where it is present,
            # it is always the first argument.
            if not isinstance(
                self.args[0],
                (
                    tuple,
                    ROOT.RDF.TH1DModel,
                    ROOT.RDF.TH2DModel,
                    ROOT.RDF.TH3DModel,
                    ROOT.RDF.THnDModel,
                    ROOT.RDF.THnSparseDModel,
                ),
            ):
                message = (
                    "Creating a histogram without a model is not supported in distributed mode. Please make sure to "
                    "specify the histogram model when rerunning the distributed RDataFrame application. For example:\n\n"
                    "\tHisto1D('mycolumn') --> Histo1D(('myhist', 'myhist', 100, 0, 10), 'mycolumn')\n\n"
                    "See the RDataFrame documentation for more details.")
                raise ValueError(message)


class VariationsFor(Action):
    """
    DistRDF.VariationsFor creates a specific node in the distributed
    RDataFrame graph. This acts as an action node.
    """
    pass


class InstantAction(Operation):
    """An instant action attached to a distributed RDataFrame graph node."""
    pass


class AsNumpy(InstantAction):
    """An 'AsNumpy' instant action attached to a distributed RDataFrame graph node."""
    pass


class Snapshot(InstantAction):
    """A 'Snapshot' instant action attached to a distributed RDataFrame graph node."""
    pass


class Transformation(Operation):
    """A trasformation attached to a distributed RDataFrame graph node."""
    pass


SUPPORTED_OPERATIONS: Dict[str, Union[Action, InstantAction, Transformation]] = {
    "Alias": Transformation,
    "AsNumpy": AsNumpy,
    "Count": Action,
    "DefaultValueFor": Transformation,
    "Define": Transformation,
    "DefinePerSample": Transformation,
    "Filter": Transformation,
    "FilterAvailable": Transformation,
    "FilterMissing": Transformation,
    "Graph": Action,
    "GraphAsymmErrors": Action,
    "Histo1D": Histo,
    "Histo2D": Histo,
    "Histo3D": Histo,
    "HistoND": Histo,
    "HistoNSparseD": Histo,
    "Max": Action,
    "Mean": Action,
    "Min": Action,
    "Profile1D": Action,
    "Profile2D": Action,
    "Profile3D": Action,
    "Redefine": Transformation,
    "Snapshot": Snapshot,
    "Stats": Action,
    "StdDev": Action,
    "Sum": Action,
    "Report": Action,
    "VariationsFor": VariationsFor,
    "Vary": Transformation,
}


def create_op(name: str, *args, **kwargs) -> Union[Action, InstantAction, Transformation]:
    try:
        return SUPPORTED_OPERATIONS[name](name, *args, **kwargs)
    except KeyError as e:
        raise ValueError(f"Operation '{name}' is either invalid or not supported in distributed mode. "
                         "See the documentation for a list of supported operations.") from e



## \file
## \ingroup tutorial_roofit_main
## \notebook -nodraw
## This macro demonstrates how to set up a fit in two ranges for plain
## likelihoods and extended likelihoods.
##
##  ### 1. Shape fits (plain likelihood)
##
##  If you fit a non-extended pdf in two ranges, e.g. `pdf.fitTo(data,Range="Range1,Range2")`,
##  it will fit the shapes in the two selected ranges and also take into account the relative
##  predicted yields in those ranges.
##
##  This is useful for example to represent a full-range fit, but with a
##  blinded signal region inside it.
##
##
##  ### 2. Shape+rate fits (extended likelihood)
##
##  If your pdf is extended, i.e. measuring both the distribution in the observable as well
##  as the event count in the fitted region, some intervention is needed to make fits in ranges
##  work in a way that corresponds to intuition.
##
##  If an extended fit is performed in a sub-range, the observed yield is only that of the subrange, hence
##  the expected event count will converge to a number that is smaller than what's visible in a plot.
##  In such cases, it is often preferred to interpret the extended term with respect to the full range
##  that's plotted, i.e., apply a correction to the extended likelihood term in such a way
##  that the interpretation of the expected event count remains that of the full range. This can
##  be done by applying a correcion factor (equal to the fraction of the pdf that is contained in the
##  fitted range) in the Poisson term that represents the extended likelihood term.
##
##  If an extended likelihood fit is performed over *two* sub-ranges, this correction is
##  even more important: without it, each component likelihood would have a different interpretation
##  of the expected event count (each corresponding to the count in its own region), and a joint
##  fit of these regions with different interpretations of the same model parameter results
##  in a number that is not easily interpreted.
##
##  If both regions correct their interpretation such that N_expected refers to the full range,
##  it is interpreted easily, and consistent in both regions.
##
##  This requires that the likelihood model is extended using RooAddPdf in the
##  form SumPdf = Nsig * sigPdf + Nbkg * bkgPdf.
##
##  \macro_image
##  \macro_code
##  \macro_output
##
## \authors Stephan Hageboeck, Wouter Verkerke

import ROOT

ROOT.gROOT.SetBatch(True)

# PART 1: Background-only fits
# ----------------------------

# Build plain exponential model
x = ROOT.RooRealVar("x", "x", 10, 100)
alpha = ROOT.RooRealVar("alpha", "alpha", -0.04, -0.1, -0.0)
model = ROOT.RooExponential("model", "Exponential model", x, alpha)

# Define side band regions and full range
x.setRange("LEFT", 10, 20)
x.setRange("RIGHT", 60, 100)

x.setRange("FULL", 10, 100)

data = model.generate(x, 10000)

# Construct an extended pdf, which measures the event count N **on the full range**.
# If the actual domain of x that is fitted is identical to FULL, this has no affect.
#
# If the fitted domain is a subset of `FULL`, though, the expected event count is divided by
# \f[
#   \mathrm{frac} = \frac{
#     \int_{\mathrm{Fit range}} \mathrm{model}(x)  \; \mathrm{d}x }{
#     \int_{\mathrm{Full range}} \mathrm{model}(x) \; \mathrm{d}x }.
# \f]
# `N` will therefore return the count extrapolated to the full range instead of the fit range.
#
# **Note**: When using a RooAddPdf for extending the likelihood, the same effect can be achieved with
# [RooAddPdf::fixCoefRange()](https://root.cern.ch/doc/master/classRooAddPdf.html#ab631caf4b59e4c4221f8967aecbf2a65),

N = ROOT.RooRealVar("N", "Extended term", 0, 20000)
extmodel = ROOT.RooExtendPdf("extmodel", "Extended model", model, N, "FULL")


# It can be instructive to fit the above model to either the LEFT or RIGHT
# range. `N` should approximately converge to the expected number of events in
# the full range. One may try to leave out `"FULL"` in the constructor, or the
# interpretation of `N` changes.
extmodel.fitTo(data, Range="LEFT", PrintLevel=-1)
N.Print()


# If we now do a simultaneous fit to the extended model, instead of the
# original model, the LEFT and RIGHT range will each correct their local `N`
# such that it refers to the `FULL` range.
#
# This joint fit of the extmodel will include (w.r.t. the plain model fit) a product of extended terms
# \f[
#     L_\mathrm{ext} = L
#       \cdot \mathrm{Poisson} \left( N_\mathrm{obs}^\mathrm{LEFT}  | N_\mathrm{exp} / \mathrm{frac LEFT} \right)
#       \cdot \mathrm{Poisson} \left( N_\mathrm{obs}^\mathrm{RIGHT} | N_\mathrm{exp} / \mathrm{frac RIGHT} \right)
# \f]


c = ROOT.TCanvas("c", "c", 2100, 700)
c.Divide(3)
c.cd(1)

r = model.fitTo(data, Range="LEFT,RIGHT", PrintLevel=-1, Save=True)
r.Print()

frame = x.frame()
data.plotOn(frame)
model.plotOn(frame, VisualizeError=r)
model.plotOn(frame)
model.paramOn(frame, Label="Non-extended fit")
frame.Draw()

c.cd(2)

r2 = extmodel.fitTo(data, Range="LEFT,RIGHT", PrintLevel=-1, Save=True)
r2.Print()
frame2 = x.frame()
data.plotOn(frame2)
extmodel.plotOn(frame2)
extmodel.plotOn(frame2, VisualizeError=r2)
extmodel.paramOn(frame2, Label="Extended fit", Layout=(0.4, 0.95))
frame2.Draw()

# PART 2: Extending with RooAddPdf
# --------------------------------
#
# Now we repeat the above exercise, but instead of explicitly adding an extended term to a single shape pdf (RooExponential),
# we assume that we already have an extended likelihood model in the form of a RooAddPdf constructed in the form `Nsig * sigPdf + Nbkg * bkgPdf`.
#
# We add a Gaussian to the previously defined exponential background.
# The signal shape parameters are chosen constant, since the signal region is entirely in the blinded region, i.e., the fit has no sensitivity.

c.cd(3)

Nsig = ROOT.RooRealVar("Nsig", "Number of signal events", 1000, 0, 2000)
Nbkg = ROOT.RooRealVar("Nbkg", "Number of background events", 10000, 0, 20000)

mean = ROOT.RooRealVar("mean", "Mean of signal model", 40.0)
width = ROOT.RooRealVar("width", "Width of signal model", 5.0)
sig = ROOT.RooGaussian("sig", "Signal model", x, mean, width)

modelsum = ROOT.RooAddPdf("modelsum", "NSig*signal + NBkg*background", [sig, model], [Nsig, Nbkg])

# This model will automatically insert the correction factor for the
# reinterpretation of Nsig and Nnbkg in the full ranges.
#
# When this happens, it reports this with lines like the following:
# ```
# [#1] INFO:Fitting -- RooAbsOptTestStatistic::ctor(nll_modelsum_modelsumData_LEFT) fixing interpretation of coefficients of any RooAddPdf to full domain of observables
# [#1] INFO:Fitting -- RooAbsOptTestStatistic::ctor(nll_modelsum_modelsumData_RIGHT) fixing interpretation of coefficients of any RooAddPdf to full domain of observables
# ```

r3 = modelsum.fitTo(data, Range="LEFT,RIGHT", PrintLevel=-1, Save=True)
r3.Print()

frame3 = x.frame()
data.plotOn(frame3)
modelsum.plotOn(frame3)
modelsum.plotOn(frame3, VisualizeError=r3)
modelsum.paramOn(frame3, Label="S+B fit with RooAddPdf", Layout=(0.3, 0.95))
frame3.Draw()

c.Draw()

c.SaveAs("rf204b_extendedLikelihood_rangedFit.png")



# \file
# \ingroup tutorial_roostats
# \notebook -js
# StandardHistFactoryPlotsWithCategories
#
#  This is a standard demo that can be used with any ROOT file
#  prepared in the standard way.  You specify:
#  - name for input ROOT file
#  - name of workspace inside ROOT file that holds model and data
#  - name of ModelConfig that specifies details for calculator tools
#  - name of dataset
#
#  With default parameters the macro will attempt to run the
#  standard hist2workspace example and read the ROOT file
#  that it produces.
#
#  The macro will scan through all the categories in a simPdf find the corresponding
#  observable.  For each category, it will loop through each of the nuisance parameters
#  and plot
#  - the data
#  - the nominal model (blue)
#  - the +Nsigma (red)
#  - the -Nsigma (green)
#
#  You can specify how many sigma to vary by changing nSigmaToVary.
#  You can also change the signal rate by changing muVal.
#
#  The script produces a lot plots, you can merge them by doing:
#
#  gs -q -dNOPAUSE -dBATCH -sDEVICE=pdfwrite -sOutputFile=merged.pdf `ls *pdf`
# ~~~
#
# \macro_image
# \macro_output
# \macro_code
#
# \author Kyle Cranmer (C++ version), and P. P. (Python translation)


import ROOT


def StandardHistFactoryPlotsWithCategories(
    infile="", workspaceName="combined", modelConfigName="ModelConfig", dataName="obsData"
):

    nSigmaToVary = 5.0
    muVal = 0
    doFit = False

    # -------------------------------------------------------
    # First part is just to access a user-defined file
    # or create the standard example file if it doesn't exist
    filename = ""
    if infile == "":
        filename = "results/example_combined_GaussExample_model.root"
        fileExist = not ROOT.gSystem.AccessPathName(filename)  # note opposite return code
        # if file does not exists generate with histfactory
        if not fileExist:
            # Normally this would be run on the command line
            print(f"will run standard hist2workspace example")
            ROOT.gROOT.ProcessLine(".! prepareHistFactory .")
            ROOT.gROOT.ProcessLine(".! hist2workspace config/example.xml")
            print(f"\n\n---------------------")
            print(f"Done creating example input")
            print(f"---------------------\n\n")

    else:
        filename = infile

    # Try to open the file
    file = ROOT.TFile.Open(filename)

    # if input file was specified but not found, quit
    if not file:
        print(f"StandardRooStatsDemoMacro: Input file {filename} is not found")
        return

    # -------------------------------------------------------
    # Tutorial starts here
    # -------------------------------------------------------

    # get the workspace out of the file
    w = file.Get(workspaceName)
    if not w:
        print(f"workspace not found")
        return

    # get the modelConfig out of the file
    mc = w.obj(modelConfigName)

    # get the modelConfig out of the file
    data = w.data(dataName)

    # make sure ingredients are found
    if not data or not mc:
        w.Print()
        print(f"data or ModelConfig was not found")
        return

    # -------------------------------------------------------
    # now use the profile inspector

    obs = mc.GetObservables().first()
    frameList = []

    firstPOI = mc.GetParametersOfInterest().first()

    firstPOI.setVal(muVal)
    #  firstPOI.setConstant()
    if doFit:
        mc.GetPdf().fitTo(data)

    # -------------------------------------------------------

    mc.GetNuisanceParameters().Print("v")
    nPlotsMax = 1000
    print(f" check expectedData by category")
    simData = ROOT.kNone
    simPdf = ROOT.kNone
    if str(mc.GetPdf().ClassName()) == "RooSimultaneous":
        print(f"Is a simultaneous PDF")
        simPdf = mc.GetPdf()
    else:
        print(f"Is not a simultaneous PDF")

    if doFit:
        channelCat = simPdf.indexCat()
        catName = channelCat.begin().first
        pdftmp = (mc.GetPdf()).getPdf(str(catName))
        obstmppdftmp.getObservables(mc.GetObservables())
        obs = obstmp.first()
        frame = obs.frame()
        print("{0}=={0}::{1}".format(channelCat.GetName(), catName))
        print(catName, " ", channelCat.getLabel())
        data.plotOn(
            frame,
            MarkerSize=1,
            Cut="{0}=={0}::{1}".format(channelCat.GetName(), catName),
            DataError="None",
        )

        normCount = data.sumEntries("{0}=={0}::{1}".format(channelCat.GetName(), catName))

        pdftmp.plotOn(frame, Normalization(normCount, ROOT.RooAbsReal.NumEvent), LineWidth=2)
        frame.Draw()
        print("expected events = ", mc.GetPdf().expectedEvents(data.get()))
        return

    nPlots = 0
    if not simPdf:

        for var in mc.GetNuisanceParameters():
            frame = obs.frame()
            frame.SetYTitle(var.GetName())
            data.plotOn(frame, MarkerSize(1))
            value = var.getVal()
            mc.GetPdf().plotOn(frame, LineWidth(1))
            var.setVal(value + var.getError());
            mc.GetPdf().plotOn(frame, LineColor(kRed), LineStyle(kDashed), LineWidth(1))
            var.setVal(value - var.getError());
            mc.GetPdf().plotOn(frame, LineColor(kGreen), LineStyle(kDashed), LineWidth(1))
            frameList.append(frame)
            var.setVal(value)

    else:
        channelCat = simPdf.indexCat()
        for tt in channelCat:

            if nPlots == nPlotsMax:
                break

            catName = tt.first

            print("on type ", catName, " ")
            # Get pdf associated with state from simpdf
            pdftmp = simPdf.getPdf(str(catName))

            # Generate observables defined by the pdf associated with this state
            obstmp = pdftmp.getObservables(mc.GetObservables())
            #      obstmp.Print()

            obs = obstmp.first()

            for var in mc.GetNuisanceParameters():

                if nPlots >= nPlotsMax:
                    break

                c2 = ROOT.TCanvas("c2")
                frame = obs.frame()
                frame.SetName(f"frame{nPlots}")
                frame.SetYTitle(var.GetName())

                cut = "{0}=={0}::{1}".format(channelCat.GetName(), catName)
                print(cut)
                print(catName, " ", channelCat.getLabel())
                data.plotOn(
                    frame,
                    MarkerSize=1,
                    Cut=cut,
                    DataError="None",
                )

                normCount = data.sumEntries(cut)

                # remember the nominal value
                value = var.getVal()

                # w.allVars().Print("v")
                # mc.GetNuisanceParameters().Print("v")
                # pdftmp.plotOn(frame,LineWidth(2))
                # mc.GetPdf().plotOn(frame,LineWidth(2.),Slice(channelCat,catName.c_str()),ProjWData(data))
                # pdftmp.plotOn(frame,LineWidth(2.),Slice(channelCat,catName.c_str()),ProjWData(data))
                global gobs
                gobs = obs
                global gpdftmp
                gpdftmp = pdftmp

                # notes: obs is RooRealVar / obstmp is RooArgSet
                #       pdftmp.expectedEvents receives RooArgSet as an argument
                #       in C++ automatic conversion is possible.
                #       in python the conversion is not possible.
                # C-code : normCount = pdftmp->expectedEvents(*obs);
                # Python : normCount = pdftmp.expectedEvents(obs) #doesn't work properly
                # RooArgSet(obs) doesn´t reproduce well the results
                # instead, we have to use obstmp
                # normCount = pdftmp.expectedEvents(RooArgSet(obstmp)) #doesn´t work properly
                normCount = pdftmp.expectedEvents(obstmp)
                pdftmp.plotOn(frame, ROOT.RooFit.Normalization(normCount, ROOT.RooAbsReal.NumEvent), LineWidth=2) # nominal

                var.setVal(value + nSigmaToVary * var.getError())

                # pdftmp.plotOn(frame,LineColor(kRed),LineStyle(kDashed),LineWidth(2))
                # mc.GetPdf().plotOn(frame,LineColor(kRed),LineStyle(kDashed),LineWidth(2.),Slice(channelCat,catName.c_str()),ProjWData(data))
                # pdftmp.plotOn(frame,LineColor(kRed),LineStyle(kDashed),LineWidth(2.),Slice(channelCat,catName.c_str()),ProjWData(data))
                normCount = pdftmp.expectedEvents(obstmp)
                pdftmp.plotOn(
                    frame,
                    ROOT.RooFit.Normalization(normCount, ROOT.RooAbsReal.NumEvent),
                    LineWidth=2,
                    LineColor="r",
                    LineStyle="--",
                ) # +n sigma

                var.setVal(value - nSigmaToVary * var.getError())

                # pdftmp.plotOn(frame,LineColor(kGreen),LineStyle(kDashed),LineWidth(2))
                # mc.GetPdf().plotOn(frame,LineColor(kGreen),LineStyle(kDashed),LineWidth(2),Slice(channelCat,catName.c_str()),ProjWData(data))
                # pdftmp.plotOn(frame,LineColor(kGreen),LineStyle(kDashed),LineWidth(2),Slice(channelCat,catName.c_str()),ProjWData(data))
                normCount = pdftmp.expectedEvents(obstmp)
                pdftmp.plotOn(
                    frame,
                    ROOT.RooFit.Normalization(normCount, ROOT.RooAbsReal.NumEvent),
                    LineWidth=2,
                    LineColor="g",
                    LineStyle="--",
                ) # -n sigma

                # set them back to normal
                var.setVal(value)

                frameList.append(frame)

                # quit making plots
                nPlots += 1

                frame.Draw()
                c2.Update()
                c2.Draw()
                c2.SaveAs(f"StandardHistFactoryPlotsWithCategories.1.{catName}_{obs.GetName()}_{var.GetName()}.png")
                del c2

    # -------------------------------------------------------

    # now make plots
    c1 = ROOT.TCanvas("c1", "ProfileInspectorDemo", 800, 200)
    nFrames = len(frameList)
    if nFrames > 4:
        nx = int(sqrt(nFrames))
        ny = ROOT.TMath.CeilNint(nFrames / nx)
        nx = ROOT.TMath.CeilNint(sqrt(nFrames))
        c1.Divide(ny, nx)
    else:
        c1.Divide(nFrames)
    for i in range(nFrames):
        c1.cd(i + 1)
        frameList[i].Draw()
        c1.Update()
    c1.Draw()
    c1.SaveAs("StandardHistFactoryPlotsWithCategories.2.pdf")

    file.Close()


StandardHistFactoryPlotsWithCategories(
    infile="", workspaceName="combined", modelConfigName="ModelConfig", dataName="obsData"
)



## \file
## \ingroup tutorial_graphs
## \notebook -js
## \preview Create a TGraphSmooth and show the usage of the interpolation function Approx.
##
## See the [TGraphSmooth documentation](https://root.cern/doc/master/classTGraphSmooth.html)
##
## \macro_image
## \macro_code
## \author Christian Stratowa, Jamie Gooding

from ctypes import c_double

import numpy as np
import ROOT

# vC1 = ROOT.TCanvas()
# # grxy = ROOT.TGraph()
# grin = ROOT.TGraph()
# grout = ROOT.TGraph()


def DrawSmooth(pad, title, xt, yt):
    vC1.cd(pad)
    vFrame = ROOT.gPad.DrawFrame(0, 0, 15, 150)
    vFrame.SetTitle(title)
    vFrame.SetTitleSize(0.2)
    vFrame.SetXTitle(xt)
    vFrame.SetYTitle(yt)
    grxy.SetMarkerColor(ROOT.kBlue)
    grxy.SetMarkerStyle(21)
    grxy.SetMarkerSize(0.5)
    grxy.Draw("P")
    grin.SetMarkerColor(ROOT.kRed)
    grin.SetMarkerStyle(5)
    grin.SetMarkerSize(0.7)
    grin.Draw("P")
    grout.DrawClone("LP")


# Test data (square)
n = 11
x = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 6.0, 6.0, 8.0, 9.0, 10.0])
y = np.array([1.0, 4.0, 9.0, 16.0, 25.0, 25.0, 36.0, 49.0, 64.0, 81.0, 100.0])

grxy = ROOT.TGraph(n, x, y)

# X values, for which y values should be interpolated
nout = 14
xout = np.array([1.2, 1.7, 2.5, 3.2, 4.4, 5.2, 5.7, 6.5, 7.6, 8.3, 9.7, 10.4, 11.3, 13])

# Create Canvas
vC1 = ROOT.TCanvas("vC1", "square", 200, 10, 700, 700)
vC1.Divide(2, 2)

# Initialize graph with data
grin = ROOT.TGraph(n, x, y)
# Interpolate at equidistant points (use mean for tied x-values)
gs = ROOT.TGraphSmooth("normal")
grout = gs.Approx(grin, "linear")
DrawSmooth(1, "Approx: ties = mean", "X-axis", "Y-axis")

# Re-initialize graph with data
# (since graph points were set to unique vales)
grin = ROOT.TGraph(n, x, y)
# Interpolate at given points xout
grout = gs.Approx(grin, "linear", 14, xout, 0, 130)
DrawSmooth(2, "Approx: ties = mean", "", "")

# Print output variables for given values xout
vNout = grout.GetN()
vXout = c_double()
vYout = c_double()
for k in range(vNout):
    grout.GetPoint(k, vXout, vYout)
    print(f"k= {k}  vXout[k]= {vXout.value}  vYout[k]= {vYout.value}")

# Re-initialize graph with data
grin = ROOT.TGraph(n, x, y)
# Interpolate at equidistant points (use min for tied x-values)
#   _grout = gs.Approx(grin,"linear", 50, 0, 0, 0, 1, 0, "min")_
grout = gs.Approx(grin, "constant", 50, 0, 0, 0, 1, 0.5, "min")
DrawSmooth(3, "Approx: ties = min", "", "")

# Re-initialize graph with data
grin = ROOT.TGraph(n, x, y)
# Interpolate at equidistant points (use max for tied x-values)
grout = gs.Approx(grin, "linear", 14, xout, 0, 0, 2, 0, "max")
DrawSmooth(4, "Approx: ties = max", "", "")



## \file
## \ingroup tutorial_hist
## \notebook -js
## Simple example illustrating how to use the C++ interpreter.
##
## \macro_image
## \macro_code
##
## \date July 2025
## \author Wim Lavrijsen, Nursena Bitirgen

import matplotlib.pyplot as plt
import mplhep as hep
import numpy as np
from ROOT import TH1F, gBenchmark, gRandom

# Create a new canvas, and enable interactive mode.
plt.ion()
fig, ax = plt.subplots(figsize=(8, 6), num="The HSUM Example")

gBenchmark.Start("hsum")

# Create some histograms.
BINS = 100
RANGE_MIN, RANGE_MAX = -4, 4
total = TH1F("total", "This is the total distribution", BINS, RANGE_MIN, RANGE_MAX)
main = TH1F("main", "Main contributor", BINS, RANGE_MIN, RANGE_MAX)
s1 = TH1F("s1", "This is the first signal", BINS, RANGE_MIN, RANGE_MAX)
s2 = TH1F("s2", "This is the second signal", BINS, RANGE_MIN, RANGE_MAX)
# total.Sumw2()

# initialize a dictionary that holds the histogram counts as numpy arrays
counts = {"total": np.zeros(BINS), "main": np.zeros(BINS), "s1": np.zeros(BINS), "s2": np.zeros(BINS)}

# Initialize random number generator.
gRandom.SetSeed()
gauss, landau = gRandom.Gaus, gRandom.Landau

# def gauss(loc, scale):
#     return np.random.normal(loc, scale)

# def landau(loc, scale):
#     return np.random.standard_cauchy() * scale + loc


# initialize the histogram filling method
def fill_hist(hist_name, x, weight=1.0):
    if RANGE_MIN <= x < RANGE_MAX:
        idx = int((x - RANGE_MIN) / (RANGE_MAX - RANGE_MIN) * BINS)
        counts[hist_name][idx] += weight


# Fill histograms randomly
kUPDATE = 500
N_EVENTS = 10000
for i in range(1, N_EVENTS + 1):
    # Generate random values.
    xmain = gauss(-1, 1.5)
    xs1 = gauss(-0.5, 0.5)
    xs2 = landau(1, 0.15)

    # Fill histograms
    # Compute the counts
    fill_hist("main", xmain)
    fill_hist("s1", xs1, 0.3)
    fill_hist("s2", xs2, 0.2)
    fill_hist("total", xmain)
    fill_hist("total", xs1, 0.3)
    fill_hist("total", xs2, 0.2)
    # Set the bin contents
    total[...] = counts["total"]
    main[...] = counts["main"]
    s1[...] = counts["s1"]
    s2[...] = counts["s2"]

    # Update display every kUPDATE events.
    if i % kUPDATE == 0:
        ax.cla()
        entries = total.GetEntries()
        mean = total.GetMean()
        stddev = total.GetStdDev()
        stats_text = f"Entries = {entries:.0f}\nMean = {mean:.2f}\nStd Dev = {stddev:.2f}"
        hep.histplot(main, histtype="fill", color="gray", alpha=0.5, edgecolor="blue", linewidth=1.5, ax=ax)
        hep.histplot(total, histtype="errorbar", color="black", ecolor="blue", linewidth=2, ax=ax)
        hep.histplot(s1, histtype="errorbar", color="blue", alpha=0.7, ecolor="blue", linewidth=2, marker="+", ax=ax)
        hep.histplot(s2, histtype="errorbar", color="blue", alpha=0.7, ecolor="blue", linewidth=2, marker="+", ax=ax)
        ax.set_title("This is the total distribution", pad=20, fontsize=14, loc="center")
        ax.text(
            0.95,
            0.90,
            stats_text,
            transform=ax.transAxes,
            ha="right",
            va="top",
            fontsize=12,
            bbox=dict(facecolor="white", edgecolor="black", boxstyle="round,pad=0.2", alpha=0.9),
        )

        # Plot formatting
        ax.set_xlim(RANGE_MIN, RANGE_MAX)
        ax.set_ylim(0, max(counts["total"]) * 1.2)
        plt.pause(0.001)

# Done, show final plot.
plt.grid(True)
plt.ioff()
plt.show()



## \file
## \ingroup tutorial_graphs
## \notebook
## Example of script showing how to divide a canvas
## into adjacent subpads + axis labels on the top and right side
## of the pads. Original tutorial by Rene Brun.
##
## See the [Divide documentation](https://root.cern/doc/master/classTPad.html#a2714ddd7ba72d5def84edc1fbaea8658)
##
## Note that the last 2 arguments in
##       c1->Divide(2,2,0,0)
## define 0 space between the pads. With this, the axis labels where the pads
## touch may be cut, as in this tutorial. To avoid this, either add some spacing
## between pads (instead of 0) or change the limits of the plot in the pad (histos
## in this tutorial). E.g. h3 could be defined as
## TH2F *h3 = new TH2F("h3","test3",10,0,1,22,-1.1,1.1);
## but note that this can change the displayed axis labels (requiring SetNdivisions
## to readjust).
##
## SetLabelOffset changes the (perpendicular) distance to the axis. The label
## position along the axis cannot be changed
## \macro_image
## \macro_code
## \author Alberto Ferro

import ROOT

c1 = ROOT.TCanvas("c1","multipads",900,700)
ROOT.gStyle.SetOptStat(0)

c1.Divide(2,2,0,0)
h1 = ROOT.TH2F("h1","test1",10,0,1,20,0,20)
h2 = ROOT.TH2F("h2","test2",10,0,1,20,0,100)
h3 = ROOT.TH2F("h3","test3",10,0,1,20,-1,1)
h4 = ROOT.TH2F("h4","test4",10,0,1,20,0,1000)

c1.cd(1)
ROOT.gPad.SetTickx(2)
h1.Draw()
c1.cd(2)
ROOT.gPad.SetTickx(2)
ROOT.gPad.SetTicky(2)
h2.GetYaxis().SetLabelOffset(0.01)
h2.Draw()
c1.cd(3)
h3.Draw()
c1.cd(4)
ROOT.gPad.SetTicky(2)
h4.Draw()




## \file
## \ingroup tutorial_vecops
## \notebook -nodraw
## This tutorial illustrates the conversion of RVec to numpy
## arrays without copying the data.
## The memory-adoption is achieved by the dictionary __array_interface__, which
## is added dynamically to the Python objects by PyROOT.
##
## \macro_code
## \macro_output
##
## \date April 2018
## \author Stefan Wunsch

import ROOT
from sys import exit

try:
    import numpy as np
except:
    exit()

# Create a vector ROOT object and assign values
# Note that this works as well with a ROOT.std.vector
vec = ROOT.RVecF(2) # or ROOT.std.vector("float")(2)
vec[0] = 1
vec[1] = 2
print("Content of the ROOT vector object: {}".format([x for x in vec]))

# Interface ROOT vector with a numpy array
array = np.asarray(vec)
print("Content of the associated numpy array: {}".format([x for x in array]))

# The numpy array adopts the memory of the vector without copying the content.
# Note that the first entry of the numpy array changes when assigning a new
# value to the first entry of the ROOT vector.
vec[0] = 42
print(
    "Content of the numpy array after changing the first entry of the ROOT vector: {}".
    format([x for x in array]))

# Use numpy features on data of ROOT objects
print("Mean of the numpy array entries: {}".format(np.mean(array)))



# Author: Enric Tejedor CERN  04/2019

################################################################################
# Copyright (C) 1995-2019, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################

import sys
import time


class PyROOTApplication(object):
    """
    Application class for PyROOT.
    Configures the interactive usage of ROOT from Python.
    """

    def __init__(self, config, is_ipython):
        from ROOT.libROOTPythonizations import InitApplication

        # Construct a TApplication for PyROOT
        InitApplication(config.IgnoreCommandLineOptions)

        self._is_ipython = is_ipython

    @staticmethod
    def _ipython_config():
        # Integrate IPython >= 5 with ROOT's event loop
        # Check for new GUI events until there is some user input to process

        from IPython import get_ipython
        from IPython.terminal import pt_inputhooks
        from IPython.terminal.interactiveshell import TerminalInteractiveShell

        def inputhook(context):
            import ROOT

            while not context.input_is_ready():
                ROOT.gSystem.ProcessEvents()
                time.sleep(0.01)

        pt_inputhooks.register("ROOT", inputhook)

        ipy = get_ipython()

        # Only the TerminalInteractiveShell will use the input hooks that are
        # registered via terminal.pt_inputhooks.
        if ipy and isinstance(ipy, TerminalInteractiveShell):
            get_ipython().run_line_magic("gui", "ROOT")

    @staticmethod
    def _inputhook_config():
        # PyOS_InputHook-based mechanism
        # Point to a function which will be called when Python's interpreter prompt
        # is about to become idle and wait for user input from the terminal
        from ROOT.libROOTPythonizations import InstallGUIEventInputHook

        InstallGUIEventInputHook()

    @staticmethod
    def _set_display_hook():
        # Set the display hook

        orig_dhook = sys.displayhook

        def displayhook(v):
            # sys.displayhook is called on the result of evaluating an expression entered
            # in an interactive Python session.
            # Therefore, this function will call EndOfLineAction after each interactive
            # command (to update display etc.)
            import ROOT

            ROOT.gInterpreter.EndOfLineAction()
            return orig_dhook(v)

        sys.displayhook = displayhook

    def init_graphics(self, gEnv, gSystem):
        """Configure ROOT graphics to be used interactively"""

        # Note that we only end up in this function if gROOT.IsBatch() is false
        import __main__

        if self._is_ipython and "IPython" in sys.modules and sys.modules["IPython"].version_info[0] >= 5:
            # ipython and notebooks, register our event processing with their hooks
            self._ipython_config()
        elif (sys.flags.interactive == 1 or not hasattr(__main__, "__file__")) and not gSystem.InheritsFrom(
            "TWinNTSystem"
        ):
            # Python in interactive mode, use the PyOS_InputHook to call our event processing
            # - sys.flags.interactive checks for the -i flags passed to python
            # - __main__ does not have the attribute __file__ if the Python prompt is started directly
            # - does not work properly on Windows
            self._inputhook_config()
            gEnv.SetValue("WebGui.ExternalProcessEvents", "yes")
        else:
            # Python in script mode, instead of separate thread methods like canvas.Update should run events

            # indicate that ProcessEvents called in different thread, let ignore thread id checks in RWebWindow
            gEnv.SetValue("WebGui.ExternalProcessEvents", "yes")

        self._set_display_hook()



## \file
## \ingroup tutorial_graphs
## \notebook
## \preview Draw a scatter plot for 4 variables, mapped to: x, y, marker colour and marker size.
##
## TScatter is available since ROOT v.6.30. See the [TScatter documentation](https://root.cern/doc/master/classTScatter.html)
##
## \macro_image
## \macro_code
## \author Olivier Couet, Jamie Gooding

import numpy as np
import ROOT

canvas = ROOT.TCanvas()
canvas.SetRightMargin(0.14)
ROOT.gStyle.SetPalette(ROOT.kBird, 0, 0.6)  # define a transparent palette

n = 175

x = np.array([])
y = np.array([])
c = np.array([])
s = np.array([])

# Define four random data sets
r = ROOT.TRandom()
for i in range(n):
    x = np.append(x, 100 * r.Rndm(i))
    y = np.append(y, 200 * r.Rndm(i))
    c = np.append(c, 300 * r.Rndm(i))
    s = np.append(s, 400 * r.Rndm(i))

scatter = ROOT.TScatter(n, x, y, c, s)
scatter.SetMarkerStyle(20)
scatter.SetTitle("Scatter plot titleX titleY titleZ title")
scatter.GetXaxis().SetRangeUser(20.0, 90.0)
scatter.GetYaxis().SetRangeUser(55.0, 90.0)
scatter.GetZaxis().SetRangeUser(10.0, 200.0)
# an alternative way to zoom the Z-axis:
# scatter->GetHistogram()->SetMinimum(10);
# scatter->GetHistogram()->SetMaximum(200);
scatter.Draw("A")



# Authors:
# * Jonas Rembser 06/2021
# * Harshal Shende 06/2021

################################################################################
# Copyright (C) 1995-2020, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################


r"""
/**
\class RooGlobalFunc
\brief \parblock \endparblock
\htmlonly
<div class="pyrootbox">
\endhtmlonly

## PyROOT

Some member functions of RooGlobalFunc that take a RooCmdArg as argument also support keyword arguments.
So far, this applies to FitOptions, Format, Frame, MultiArg, YVar and ZVar.
\code{.py}
# Directly passing a RooCmdArg:
ROOT.RooMCStudy(model, ROOT.RooArgSet(x), ROOT.RooFit.FitOptions=(ROOT.RooFit.Save(True), ROOT.RooFit.PrintEvalErrors(0)))

# With keyword arguments:
ROOT.RooMCStudy(model, ROOT.RooArgSet(x), FitOptions=dict(Save=True, PrintEvalErrors=0))
\endcode

\htmlonly
</div>
\endhtmlonly
*/
"""

from ._utils import _dict_to_flat_map, _kwargs_to_roocmdargs, cpp_signature


@cpp_signature(
    "RooFit::FitOptions(const RooCmdArg& arg1, const RooCmdArg& arg2={},"
    "const RooCmdArg& arg3={},const RooCmdArg& arg4={},"
    "const RooCmdArg& arg5={},const RooCmdArg& arg6={}) ;"
)
def FitOptions(*args, **kwargs):
    r"""The FitOptions() function is pythonized with the command argument pythonization.
    The keywords must correspond to the CmdArg of the function.
    """
    # Redefinition of `FitOptions` for keyword arguments.
    from ROOT import RooFit

    args, kwargs = _kwargs_to_roocmdargs(*args, **kwargs)
    return RooFit._FitOptions(*args, **kwargs)


@cpp_signature(
    "RooFit::Format(const char* what, const RooCmdArg& arg1={}, const RooCmdArg& arg2={},"
    "const RooCmdArg& arg3={},const RooCmdArg& arg4={},"
    "const RooCmdArg& arg5={},const RooCmdArg& arg6={},"
    "const RooCmdArg& arg7={},const RooCmdArg& arg8={}) ;"
)
def Format(*args, **kwargs):
    r"""The Format() function is pythonized with the command argument pythonization.
    The keywords must correspond to the CmdArg of the function.
    """
    # Redefinition of `Format` for keyword arguments.
    from ROOT import RooFit

    if "what" in kwargs:
        args = (kwargs["what"],) + args
        del kwargs["what"]
    args, kwargs = _kwargs_to_roocmdargs(*args, **kwargs)
    return RooFit._Format(*args, **kwargs)


@cpp_signature(
    "RooFit::Frame(const RooCmdArg& arg1, const RooCmdArg& arg2={},"
    "const RooCmdArg& arg3={}, const RooCmdArg& arg4={},"
    "const RooCmdArg& arg5={}, const RooCmdArg& arg6={}) ;"
)
def Frame(*args, **kwargs):
    r"""The Frame() function is pythonized with the command argument pythonization.
    The keywords must correspond to the CmdArg of the function.
    """
    # Redefinition of `Frame` for keyword arguments.
    from ROOT import RooFit

    args, kwargs = _kwargs_to_roocmdargs(*args, **kwargs)
    return RooFit._Frame(*args, **kwargs)


@cpp_signature(
    "RooFit::MultiArg(const RooCmdArg& arg1, const RooCmdArg& arg2,"
    "const RooCmdArg& arg3={},const RooCmdArg& arg4={},"
    "const RooCmdArg& arg5={},const RooCmdArg& arg6={},"
    "const RooCmdArg& arg7={},const RooCmdArg& arg8={}) ;"
)
def MultiArg(*args, **kwargs):
    r"""The MultiArg() function is pythonized with the command argument pythonization.
    The keywords must correspond to the CmdArg of the function.
    """
    # Redefinition of `MultiArg` for keyword arguments.
    from ROOT import RooFit

    args, kwargs = _kwargs_to_roocmdargs(*args, **kwargs)
    return RooFit._MultiArg(*args, **kwargs)


@cpp_signature("RooFit::YVar(const RooAbsRealLValue& var, const RooCmdArg& arg={}) ;")
def YVar(*args, **kwargs):
    r"""The YVar() function is pythonized with the command argument pythonization.
    The keywords must correspond to the CmdArg of the function.
    """
    # Redefinition of `YVar` for keyword arguments.
    from ROOT import RooFit

    if "var" in kwargs:
        args = (kwargs["var"],) + args
        del kwargs["var"]
    args, kwargs = _kwargs_to_roocmdargs(*args, **kwargs)
    return RooFit._YVar(*args, **kwargs)


@cpp_signature("RooFit::ZVar(const RooAbsRealLValue& var, const RooCmdArg& arg={}) ;")
def ZVar(*args, **kwargs):
    r"""The ZVar() function is pythonized with the command argument pythonization.
    The keywords must correspond to the CmdArg of the function.
    """
    # Redefinition of `ZVar` for keyword arguments.
    from ROOT import RooFit

    if "var" in kwargs:
        args = (kwargs["var"],) + args
        del kwargs["var"]
    args, kwargs = _kwargs_to_roocmdargs(*args, **kwargs)
    return RooFit._ZVar(*args, **kwargs)


@cpp_signature("RooFit::Slice(std::map<RooCategory*, std::string> const&) ;")
def Slice(*args, **kwargs):
    r"""The Slice function is pythonized for converting python dict to std::map.
    The keywords must correspond to the CmdArg of the function.
    The instances in the dict must correspond to the template argument in std::map of the function.
    """
    # Redefinition of `Slice` for keyword arguments and converting python dict to std::map.
    from ROOT import RooFit

    if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], dict):
        return RooFit.Detail.SliceFlatMap(_dict_to_flat_map(args[0], {"RooCategory*": "std::string"}))

    return RooFit._Slice(*args, **kwargs)


@cpp_signature(
    [
        "RooFit::Import(const std::map<std::string,RooDataSet*>& ) ;",
        "RooFit::Import(const std::map<std::string,TH1*>&) ;",
        "RooFit::Import(const std::map<std::string,RooDataHist*>&) ;",
    ]
)
def Import(*args, **kwargs):
    r"""The Import function is pythonized for converting python dict to std::map.
    The keywords must correspond to the CmdArg of the function.
    The instances in the dict must correspond to the template argument in std::map of the function.
    """
    # Redefinition of `Import` for keyword arguments and converting python dict to std::map.
    from ROOT import RooFit

    if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], dict):
        return RooFit.Detail.ImportFlatMap(
            _dict_to_flat_map(args[0], {"std::string": ["TH1*", "RooDataHist*", "RooDataSet*"]})
        )

    return RooFit._Import(*args, **kwargs)


@cpp_signature("RooFit::Link(const std::map<std::string,RooAbsData*>&) ;")
def Link(*args, **kwargs):
    r"""The Link function is pythonized for converting python dict to std::map.
    The keywords must correspond to the CmdArg of the function.
    The instances in the dict must correspond to the template argument in std::map of the function.
    """
    # Redefinition of `Link` for keyword arguments and converting python dict to std::map.
    from ROOT import RooFit

    if len(args) == 1 and len(kwargs) == 0 and isinstance(args[0], dict):
        return RooFit.Detail.LinkFlatMap(_dict_to_flat_map(args[0], {"std::string": "RooAbsData*"}))

    return RooFit._Link(*args, **kwargs)


@cpp_signature("RooFit::DataError(Int_t) ;")
def DataError(etype):
    r"""If you want to use the `"None"` enum value to disable error plotting, you
    can also pass `None` directly instead of passing a string:

    ~~~ {.py}
        data.plotOn(frame, DataError=None)
        # instead of DataError="None"
    ~~~
    """
    from ROOT import RooFit

    # One of the possible enum values is "None", and we want the user to be
    # able to pass None also as a NoneType for convenience.
    if etype is None:
        etype = "None"

    return RooFit._DataError(etype)


def _bindFunctionOrPdf(name, func, is_rooabspdf, *variables):
    """
    Wrap an arbitrary function defined in Python or C++.

    If you're wrapping a Python function, it must take numpy arrays of type
    float64 as input and output types.

    Parameters:
    - name (str): Name of the function.
    - func (callable): Function that defines the function.
    - variables (list): List of variables to be used in the function.

    Returns:
    - RooAbsReal wrapping the given function
    """

    import ROOT

    # use the C++ version if dealing with C++ function
    if "cppyy" in repr(type(func)):
        return ROOT.RooFit._bindFunction(name, func, *variables)

    base_class_name = "RooAbsPdf" if is_rooabspdf else "RooAbsReal"

    class RooPyBindDerived(ROOT.RooFit.Detail.RooPyBind[base_class_name]):

        _outputBuffer = None

        def __init__(self, name, title, *variables):
            super(RooPyBindDerived, self).__init__(name, title, ROOT.RooArgList(*variables))

        def evaluate(self):
            import numpy as np

            inputs = [np.array([v.getVal()]) for v in self.varlist()]
            return func(*inputs)[0]

        def doEvalPy(self, ctx):
            import numpy as np

            def span_to_numpy(sp):
                return np.frombuffer(sp.data(), dtype=np.float64, count=sp.size())

            inputs = [span_to_numpy(ctx.at(v)) for v in self.varlist()]
            if self._outputBuffer is None:
                self._outputBuffer = np.zeros(ctx.output().size())
            self._outputBuffer[:] = func(*inputs)
            return self._outputBuffer

        def clone(self, newname=False):
            cl = RooPyBindDerived(newname if newname else self.GetName(), self.GetTitle(), self.varlist())
            ROOT.SetOwnership(cl, False)
            return cl

    return RooPyBindDerived(name, "", variables)


def bindFunction(name, func, *variables):
    return _bindFunctionOrPdf(name, func, False, *variables)


def bindPdf(name, func, *variables):
    return _bindFunctionOrPdf(name, func, True, *variables)



## \file
## \ingroup tutorial_hist
## \notebook
## Example of macro illustrating how to superimpose two histograms
## with different scales in the "same" pad.
## Inspired by work of Rene Brun.
##
## \macro_image
## \macro_code
##
## \date July 2025
## \author Alberto Ferro, Nursena Bitirgen

import matplotlib.pyplot as plt
import mplhep as hep
import numpy as np
import ROOT

np.random.seed(0)
plt.style.use(hep.style.ROOT)
fig, ax1 = plt.subplots(figsize=(10, 6))

h1 = ROOT.TH1F("h1", "my histogram", 100, -3, 3)

h1[...] = np.histogram(np.random.normal(0, 1, 10000), range=(-3, 3), bins=100)[0]

hep.histplot(h1, ax=ax1, histtype="fill", color="white", edgecolor="blue", linewidth=1.5, alpha=0.5)

hint1 = ROOT.TH1F("hint1", "h1 bins integral", 100, -3, 3)

hint1[...] = np.cumsum(h1.values())

ax2 = ax1.twinx()
hep.histplot(hint1, ax=ax2, histtype="errorbar", color="red", marker="+", markersize=3)

ax1.set_xlim(-3, 3)
plt.title("Histogram with Cumulative Sum")
plt.show()



## \file
## \ingroup tutorial_dataframe
## \notebook
## Read data from Pandas Data Frame into RDataFrame.
##
## \macro_code
## \macro_output
##
## \date February 2024
## \author Pere Mato (CERN)

import ROOT
import pandas as pd

# Let's create some data in a pandas dataframe
pdf = pd.DataFrame({'x': [1, 2, 3], 'y': [4, 5, 6]})

# Convert the Pandas DataFrame to RDataFrame
# The column names are directly copied to the RDF 
# Please note that only fundamental types (int, float, ...) are supported and
# the arrays must have the same length.
df = ROOT.RDF.FromPandas(pdf)

# You can now use the RDataFrame as usually, e.g. add a column ...
df = df.Define('z', 'x + y')

# ... or print the content
df.Display().Print()

# ... or save the data as a ROOT file
df.Snapshot('tree', 'df035_RDFFromPandas.root')



# Authors:
# * Jonas Rembser 06/2021
# * Harshal Shende 06/2021

################################################################################
# Copyright (C) 1995-2020, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################


from ._utils import _dict_to_flat_map, _kwargs_to_roocmdargs, cpp_signature


class RooSimultaneous(object):
    r"""Some member functions of RooSimultaneous that take a RooCmdArg as argument also support keyword arguments.
    So far, this applies to RooSimultaneous::plotOn.
    For example, the following code is equivalent in PyROOT:
    \code{.py}
    # Directly passing a RooCmdArg:
    pdfSim.plotOn(frame, ROOT.RooFit.Slice(sample,"control"), ROOT.RooFit.ProjWData(sampleSet, combData))

    # With keyword arguments:
    simPdf.plotOn(frame, Slice=(sample, "control"), ProjWData=(sampleSet, combData))
    \endcode
    """

    __cpp_name__ = 'RooSimultaneous'

    @cpp_signature(
        "RooSimultaneous(const char *name, const char *title,"
        "                std::map<std::string,RooAbsPdf*> pdfMap, RooAbsCategoryLValue& inIndexCat) ;"
    )
    def __init__(self, *args):
        r"""The RooSimultaneous constructor that takes a map of category names
        to PDFs is accepting a Python dictionary in Python.
        """
        if len(args) >= 3 and isinstance(args[2], dict):
            args = list(args)
            args[2] = _dict_to_flat_map(args[2], {"std::string": "RooAbsPdf*"})
        self._init(*args)

    @cpp_signature(
        "RooPlot *RooSimultaneous::plotOn(RooPlot* frame,"
        "    const RooCmdArg& arg1            , const RooCmdArg& arg2={},"
        "    const RooCmdArg& arg3={}, const RooCmdArg& arg4={},"
        "    const RooCmdArg& arg5={}, const RooCmdArg& arg6={},"
        "    const RooCmdArg& arg7={}, const RooCmdArg& arg8={},"
        "    const RooCmdArg& arg9={}, const RooCmdArg& arg10={}) const;"
    )
    def plotOn(self, *args, **kwargs):
        r"""The RooSimultaneous::plotOn() function is pythonized with the command argument pythonization.
        The keywords must correspond to the CmdArgs of the function.
        """
        # Redefinition of `RooSimultaneous.plotOn` for keyword arguments.
        args, kwargs = _kwargs_to_roocmdargs(*args, **kwargs)
        return self._plotOn(*args, **kwargs)



# Author: Vincenzo Eduardo Padulano CERN  10/2023

################################################################################
# Copyright (C) 1995-2023, Rene Brun and Fons Rademakers.                      #
# All rights reserved.                                                         #
#                                                                              #
# For the licensing terms see $ROOTSYS/LICENSE.                                #
# For the list of contributors see $ROOTSYS/README/CREDITS.                    #
################################################################################

from . import pythonization

# TODO: vpadulan
# This module enables pickling/unpickling of the std::runtime_error Python proxy
# defined in cppyy (via CPPExcInstance). The same logic should be implemented
# in the CPython extension to be more generic.

def _create_error(what):
    """
    Creates a new std::runtime_error proxy with the original message.
    """
    import ROOT

    return ROOT.std.runtime_error(what)

def _reduce_error(self):
    """
    Establish the strategy to recreate a std::runtime_error when unpickling.

    Creates a new error from the original message. Need to use a separate free
    function instead of an inline lambda to help pickle.
    """
    return _create_error, (self.what(), )

@pythonization("runtime_error", ns="std")
def pythonize_runtime_error(klass):
    """Add serialization capabilities to std::runtime_error."""
    klass.__reduce__ = _reduce_error



## \file
## \ingroup tutorial_hist_legacy
## Display two histograms and their ratio.
##
## This program illustrates how to plot two histograms and their
## ratio on the same canvas. Original macro by Olivier Couet.
##
## \macro_code
##
## \author Michael Moran

from ROOT import TCanvas, TColor, TGaxis, TH1F, TPad
from ROOT import kBlack, kBlue, kRed


def createH1():
    h1 = TH1F("h1", ("Two gaussian plots and their ratio; x title; h1 and h2"
                " histograms"), 100, -5, 5)
    h1.SetLineColor(kBlue+1)
    h1.SetLineWidth(2)
    h1.FillRandom("gaus")
    h1.GetYaxis().SetTitleSize(20)
    h1.GetYaxis().SetTitleFont(43)
    h1.GetYaxis().SetTitleOffset(1.55)
    h1.SetStats(0)
    return h1


def createH2():
    h2 = TH1F("h2", "h2", 100, -5, 5)
    h2.FillRandom("gaus")
    h2.SetLineColor(kRed)
    h2.SetLineWidth(2)
    return h2


def createRatio(h1, h2):
    h3 = h1.Clone("h3")
    h3.SetLineColor(kBlack)
    h3.SetMarkerStyle(21)
    h3.SetTitle("")
    h3.SetMinimum(0.8)
    h3.SetMaximum(1.35)
    # Set up plot for markers and errors
    h3.Sumw2()
    h3.SetStats(0)
    h3.Divide(h2)

    # Adjust y-axis settings
    y = h3.GetYaxis()
    y.SetTitle("ratio h1/h2 ")
    y.SetNdivisions(505)
    y.SetTitleSize(20)
    y.SetTitleFont(43)
    y.SetTitleOffset(1.55)
    y.SetLabelFont(43)
    y.SetLabelSize(15)

    # Adjust x-axis settings
    x = h3.GetXaxis()
    x.SetTitleSize(20)
    x.SetTitleFont(43)
    x.SetTitleOffset(4.0)
    x.SetLabelFont(43)
    x.SetLabelSize(15)

    return h3


def createCanvasPads():
    c = TCanvas("c", "canvas", 800, 800)
    # Upper histogram plot is pad1
    pad1 = TPad("pad1", "pad1", 0, 0.3, 1, 1.0)
    pad1.SetBottomMargin(0)  # joins upper and lower plot
    pad1.SetGridx()
    pad1.Draw()
    # Lower ratio plot is pad2
    c.cd()  # returns to main canvas before defining pad2
    pad2 = TPad("pad2", "pad2", 0, 0.05, 1, 0.3)
    pad2.SetTopMargin(0)  # joins upper and lower plot
    pad2.SetBottomMargin(0.2)
    pad2.SetGridx()
    pad2.Draw()

    return c, pad1, pad2


def ratioplot():
    # create required parts
    h1 = createH1()
    h2 = createH2()
    h3 = createRatio(h1, h2)
    c, pad1, pad2 = createCanvasPads()

    # draw everything
    pad1.cd()
    h1.Draw()
    h2.Draw("same")
    # to avoid clipping the bottom zero, redraw a small axis
    h1.GetYaxis().SetLabelSize(0.0)
    axis = TGaxis(-5, 20, -5, 220, 20, 220, 510, "")
    axis.SetLabelFont(43)
    axis.SetLabelSize(15)
    axis.Draw()
    pad2.cd()
    h3.Draw("ep")

# To hold window open when running from command line
#  text = raw_input()


if __name__ == "__main__":
    ratioplot()



import math

from .. import get_keras_version


def MakeKerasConv(layer):
    """
    Create a Keras-compatible convolutional layer operation using SOFIE framework.

    This function takes a dictionary representing a convolutional layer and its attributes and
    constructs a Keras-compatible convolutional layer operation using the SOFIE framework.
    A convolutional layer applies a convolution operation between the input tensor and a set
    of learnable filters (kernels).

    Parameters:
    layer (dict): A dictionary containing layer information including input, output,
                  data type (must be float), weight and bias name, kernel size, dilations, padding and strides.
                  When padding is same (keep in the same dimensions), the padding shape is calculated.

    Returns:
    ROperator_Conv: A SOFIE framework operator representing the convolutional layer operation.
    """
    from ROOT.TMVA.Experimental import SOFIE

    keras_version = get_keras_version()

    finput = layer["layerInput"]
    foutput = layer["layerOutput"]
    fLayerDType = layer["layerDType"]
    fLayerInputName = finput[0]
    fLayerOutputName = foutput[0]
    attributes = layer["layerAttributes"]
    fWeightNames = layer["layerWeight"]
    fKernelName = fWeightNames[0]
    fBiasName = fWeightNames[1]
    fAttrDilations = attributes["dilation_rate"]
    fAttrGroup = int(attributes["groups"])
    fAttrKernelShape = attributes["kernel_size"]
    fKerasPadding = str(attributes["padding"])
    fAttrStrides = attributes["strides"]
    fAttrPads = []

    if fKerasPadding == "valid":
        fAttrAutopad = "VALID"
    elif fKerasPadding == "same":
        fAttrAutopad = "NOTSET"
        if keras_version < "2.16":
            fInputShape = attributes["_build_input_shape"]
        else:
            fInputShape = attributes["_build_shapes_dict"]["input_shape"]
        if layer.get("channels_last", True):
            inputHeight = fInputShape[1]
            inputWidth = fInputShape[2]
        else:
            inputHeight = fInputShape[2]
            inputWidth = fInputShape[3]
        outputHeight = math.ceil(float(inputHeight) / float(fAttrStrides[0]))
        outputWidth = math.ceil(float(inputWidth) / float(fAttrStrides[1]))
        padding_height = max((outputHeight - 1) * fAttrStrides[0] + fAttrKernelShape[0] - inputHeight, 0)
        padding_width = max((outputWidth - 1) * fAttrStrides[1] + fAttrKernelShape[1] - inputWidth, 0)
        padding_top = math.floor(padding_height / 2)
        padding_bottom = padding_height - padding_top
        padding_left = math.floor(padding_width / 2)
        padding_right = padding_width - padding_left
        fAttrPads = [padding_top, padding_bottom, padding_left, padding_right]
    else:
        raise RuntimeError(
            "TMVA::SOFIE - RModel Keras Parser doesn't yet supports Convolution layer with padding " + fKerasPadding
        )
    if SOFIE.ConvertStringToType(fLayerDType) == SOFIE.ETensorType.FLOAT:
        op = SOFIE.ROperator_Conv["float"](
            fAttrAutopad,
            fAttrDilations,
            fAttrGroup,
            fAttrKernelShape,
            fAttrPads,
            fAttrStrides,
            fLayerInputName,
            fKernelName,
            fBiasName,
            fLayerOutputName,
        )
        return op
    else:
        raise RuntimeError("TMVA::SOFIE - Unsupported - Operator Conv does not yet support input type " + fLayerDType)



## \file
## \ingroup tutorial_roofit_main
## \notebook
## Multidimensional models: usage of full pdf with per-event errors
##
## \macro_code
##
## \date February 2018
## \authors Clemens Lange, Wouter Verkerke (C++ version)

import ROOT

# B-physics pdf with per-event Gaussian resolution
# ----------------------------------------------------------------------------------------------

# Observables
dt = ROOT.RooRealVar("dt", "dt", -10, 10)
dterr = ROOT.RooRealVar("dterr", "per-event error on dt", 0.01, 10)

# Build a gaussian resolution model scaled by the per-error =
# gauss(dt,bias,sigma*dterr)
bias = ROOT.RooRealVar("bias", "bias", 0, -10, 10)
sigma = ROOT.RooRealVar("sigma", "per-event error scale factor", 1, 0.1, 10)
gm = ROOT.RooGaussModel("gm1", "gauss model scaled bt per-event error", dt, bias, sigma, dterr)

# Construct decay(dt) (x) gauss1(dt|dterr)
tau = ROOT.RooRealVar("tau", "tau", 1.548)
decay_gm = ROOT.RooDecay("decay_gm", "decay", dt, tau, gm, type="DoubleSided")

# Construct empirical pdf for per-event error
# -----------------------------------------------------------------

# Use landau pdf to get empirical distribution with long tail
pdfDtErr = ROOT.RooLandau("pdfDtErr", "pdfDtErr", dterr, 1.0, 0.25)
expDataDterr = pdfDtErr.generate({dterr}, 10000)

# Construct a histogram pdf to describe the shape of the dtErr distribution
expHistDterr = expDataDterr.binnedClone()
pdfErr = ROOT.RooHistPdf("pdfErr", "pdfErr", {dterr}, expHistDterr)

# Construct conditional product decay_dm(dt|dterr)*pdf(dterr)
# ----------------------------------------------------------------------------------------------------------------------

# Construct production of conditional decay_dm(dt|dterr) with empirical
# pdfErr(dterr)
model = ROOT.RooProdPdf("model", "model", {pdfErr}, Conditional=({decay_gm}, {dt}))

# (Alternatively you could also use the landau shape pdfDtErr)
# ROOT.RooProdPdf model("model", "model",pdfDtErr,
# ROOT.RooFit.Conditional(decay_gm,dt))

# Sample, fit and plot product model
# ------------------------------------------------------------------

# Specify external dataset with dterr values to use model_dm as
# conditional pdf
data = model.generate({dt, dterr}, 10000)

# Fit conditional decay_dm(dt|dterr)
# ---------------------------------------------------------------------

# Specify dterr as conditional observable
model.fitTo(data, PrintLevel=-1)

# Plot conditional decay_dm(dt|dterr)
# ---------------------------------------------------------------------

# Make two-dimensional plot of conditional pdf in (dt,dterr)
hh_model = model.createHistogram("hh_model", dt, Binning=50, YVar=dict(var=dterr, Binning=50))
hh_model.SetLineColor("kBlue")

# Make projection of data an dt
frame = dt.frame(Title="Projection of model(dt|dterr) on dt")
data.plotOn(frame)
model.plotOn(frame)

# Draw all frames on canvas
c = ROOT.TCanvas("rf307_fullpereventerrors", "rf307_fullpereventerrors", 800, 400)
c.Divide(2)
c.cd(1)
ROOT.gPad.SetLeftMargin(0.20)
hh_model.GetZaxis().SetTitleOffset(2.5)
hh_model.Draw("surf")
c.cd(2)
ROOT.gPad.SetLeftMargin(0.15)
frame.GetYaxis().SetTitleOffset(1.6)
frame.Draw()

c.SaveAs("rf307_fullpereventerrors.png")



from __future__ import annotations

import logging
import uuid
import warnings
from abc import ABC, abstractmethod
from collections import Counter, deque
from copy import deepcopy
from dataclasses import dataclass
from functools import partial, singledispatch
from typing import TYPE_CHECKING, Callable, Deque, Dict, Iterable, List, Optional, Union

import ROOT

from . import ComputationGraphGenerator, Ranges, _graph_cache
from .Backends import Utils
from .Backends.Base import distrdf_mapper, distrdf_reducer
from .Node import Node
from .Operation import Action, InstantAction, Operation

if TYPE_CHECKING:
    from .Backends.Base import BaseBackend, TaskResult

logger = logging.getLogger(__name__)


@singledispatch
def _append_node_to_actions(operation: Operation, node: Node, actions: List[Node]) -> None:
    """
    Appends the input node to the list of action nodes, if the operation of the
    node is an action.
    """
    pass


@_append_node_to_actions.register(Action)
@_append_node_to_actions.register(InstantAction)
def _(operation: Union[Action, InstantAction], node: Node, actions: List[Node]) -> None:
    actions.append(node)


@dataclass
class TaskObjects:
    """
    Holds objects needed in a distributed task.
    Attributes:
        rdf: The starting node of the RDataFrame computation graph. Only in a
            TTree-based run, if the task has nothing to process then this
            attribute is None.
        entries_in_trees: A struct holding the amount of processed entries in
            the task, as well as a dictionary where each key is an identifier
            for a tree opened in the task and the value is the number of entries
            in that tree. This attribute is not None only in a TTree-based run.
    """
    rdf: Optional[ROOT.RDF.RNode]
    entries_in_trees: Optional[Ranges.TaskTreeEntries]


class HeadNode(Node, ABC):
    """
    The head node of the computation graph. Keeps record of all nodes in the
    graph, is then able to generate a flat representation to be sent to the
    distributed workers.

    Attributes:
        backend: A reference to the instance of distributed backend that will
            execute this computation graph.

        graph_nodes: A deque of references to the nodes belonging to this graph.

        node_counter: A counter of how many nodes were created in the graph of
            this head node, starting from zero.
    """

    def __init__(self, backend: BaseBackend, npartitions: Optional[int], localdf: ROOT.RDataFrame):
        super().__init__(lambda: self)

        self.backend = backend

        self.node_counter: int = 0
        # It is important to have a double-ended queue because we need to
        # traverse the nodes in different ways w.r.t. their insertion order.
        # While pruning the graph, we need to check leaf nodes before their
        # parents, so if a child is pruned then it also decrements the counter
        # of children of its parent. Thus, we need a bottom-up traversal.
        # While executing the graph in a task, we need to create the RDF nodes
        # in the order the user requested them, e.g. starting from the
        # RDataFrame itself, then calling its direct children, their children
        # and so on. Thus, we need a top-down traversal.
        self.graph_nodes: Deque[Node] = deque([self])

        # Uniquely identify each computation graph execution of this RDataFrame
        # The uuid can be set here
        self.rdf_uuid = uuid.uuid4()
        # The full identifier is created at the beginning of each execution
        self.exec_id: _graph_cache.ExecutionIdentifier = None

        # Internal attribute to keep track of the number of partitions. We also
        # check whether it was specified by the user when creating the dataframe.
        # If so, this attribute will not be updated when triggering.
        self._npartitions = npartitions
        self._user_specified_npartitions = True if npartitions is not None else False

        # Internal RDataFrame object, useful to expose information such as
        # column names.
        self.rdf_node = localdf

        # A dictionary where the keys are the IDs of the objects to live visualize
        # and the values are the corresponding callback functions 
        # This attribute only gets set in case the LiveVisualize() function is called
        self.drawables_dict: Optional[Dict[int, List[Optional[Callable]]]] = None

    def __del__(self):
        """
        Remove the reference to the local RDataFrame object as soon as this
        object is garbage collected. This helps avoiding conflicts between
        the garbage collector, the cppyy memory regulator and the C++ object
        destructor.
        """
        if hasattr(self, "rdf_node"):
            del self.rdf_node

    @property
    def npartitions(self) -> Optional[int]:
        return self._npartitions

    @npartitions.setter
    def npartitions(self, value: int) -> None:
        """
        The number of partitions for this dataframe is updated only if the user
        did not initially specify one when creating the dataframe.
        """
        if not self._user_specified_npartitions:
            self._npartitions = value

    def _prune_graph(self):
        """
        Prunes nodes from the graph under certain conditions. A node is pruned
        if it has no children and the user has no references to it. The internal
        representation of the graph is traversed in such a way so that leaf
        nodes are checked before their parents.
        """
        logger.debug("Starting computational graph pruning")
        self.graph_nodes = deque(node for node in self.graph_nodes if not node.is_prunable())
        logger.debug("Ended computational graph pruning")

    def _get_action_nodes(self) -> List[Node]:
        """Generates a list of nodes in the graph that are actions."""
        # This function is called after distributed execution of the graph, no
        # need to repeat the pruning
        action_nodes: List[Node] = []
        for node in reversed(self.graph_nodes):
            _append_node_to_actions(node.operation, node, action_nodes)
        return action_nodes

    def _generate_graph_dict(self) -> Dict[int, Node]:
        """
        Generates a dictionary holding information about all nodes in the graph.
        It is then given to the distributed scheduler.
        """
        # Need to prune the graph before sending it for distributed execution
        self._prune_graph()
        # We need to store references of each node id separately from the node
        # objects themselves. In a distributed task, we need to know for any
        # node which node is its parent (in order to execute an RDF operation
        # on the right node). We can't serialize the reference to the parent
        # node that each node object has, since that would trigger recursive
        # serialization of the parent(s) of parent(s) nodes.
        return {node.node_id: node for node in reversed(self.graph_nodes)}

    @abstractmethod
    def _build_ranges(self) -> List[Ranges.DataRange]:
        pass

    @abstractmethod
    def _generate_rdf_creator(self) -> Callable[[Ranges.DataRange], TaskObjects]:
        pass

    @abstractmethod
    def _handle_returned_values(self, values: TaskResult) -> Iterable:
        pass

    def _execute_and_retrieve_results(self, mapper, local_nodes) -> TaskResult:
        if self.drawables_dict is not None:
            # Prepare a dictionary with additional information for live visualization
            drawables_info_dict = {
                # Key: node_id
                node.node_id: (
                    # Tuple containing:
                    # 1. Callback functions passed by the user
                    self.drawables_dict[node.node_id],
                    # 2. Index of the node in the local_nodes list
                    i,
                    # 3. Name of the operation associated with the node
                    node.operation.name
                )
                for i, node in enumerate(local_nodes)
                # Filter: Only include nodes requested by the user
                if node.node_id in self.drawables_dict
            }
            return self.backend.ProcessAndMergeLive(self._build_ranges(), mapper, distrdf_reducer, drawables_info_dict)
        else:
            return self.backend.ProcessAndMerge(self._build_ranges(), mapper, distrdf_reducer)

    def execute_graph(self) -> None:
        """
        Executes an RDataFrame computation graph on a distributed backend.

        The needed ingredients are:

        - A collection of logical ranges in which the dataset is split. Each
          range is going to be assigned to a distributed task.
        - A representation of the computation graph that the task needs to
          execute.
        - A way to generate an RDataFrame instance starting from the logical
          range of the task.
        - Optionally, some setup code to be run at the beginning of each task.

        These are used as inputs to a generic mapper function. Results from the
        various mappers are then reduced and the final results are retrieved in
        the local session. These are properly handled to perform extra checks,
        depending on the data source. Finally, the local user-facing nodes are
        filled with the values that were computed distributedly so that they
        can be accessed in the application like with local RDataFrame.
        """

        # Updates the number of partitions for this dataframe if the user did
        # not specify one initially. This is done each time the computations are
        # triggered, in case the user changed the resource configuration
        # between runs (e.g. changing the number of available cores).
        self.npartitions = self.backend.optimize_npartitions()

        self.exec_id = _graph_cache.ExecutionIdentifier(self.rdf_uuid, uuid.uuid4())


        computation_graph_callable = partial(ComputationGraphGenerator.trigger_computation_graph, self._generate_graph_dict())

        # Accumulate all code that needs to be declared in one string
        code_to_declare = "\n".join(self.backend.strings_to_declare.values())

        mapper = partial(distrdf_mapper,
                         build_rdf_from_range=self._generate_rdf_creator(),
                         computation_graph_callable=computation_graph_callable,
                         initialization_fn=self.backend.initialization,
                         code_to_declare=code_to_declare)

        # List of action nodes in the same order as values
        local_nodes = self._get_action_nodes()

        # Execute graph distributedly and return the aggregated results from all tasks
        # using the appropriate backend method based on whether or not live visualization is enabled
        try:
            returned_values = self._execute_and_retrieve_results(mapper, local_nodes)
        finally:
            # Cleanup the current execution artifacts from the caches on the workers
            self.backend.cleanup_cache(self.exec_id)

        # Perform any extra checks that may be needed according to the
        # type of the head node
        final_values = self._handle_returned_values(returned_values)
        
        # Set the value of every action node
        for node, value in zip(local_nodes, final_values):
            Utils.set_value_on_node(value, node, self.backend)


def get_headnode(backend: BaseBackend, npartitions: int, *args) -> HeadNode:
    """
    A factory for different kinds of head nodes of the RDataFrame computation
    graph, depending on the arguments to the RDataFrame constructor. Parses the
    arguments and compares them against the possible RDataFrame constructors.
    """

    # Early check that arguments are accepted by RDataFrame
    try:
        localdf = ROOT.RDataFrame(*args)
    except TypeError:
        raise TypeError(("The arguments provided are not accepted by any RDataFrame constructor. "
                         "See the RDataFrame documentation for the accepted constructor argument types."))

    if isinstance(args[0], ROOT.RDF.Experimental.RDatasetSpec):
        return RDatasetSpecHeadNode(backend, npartitions, localdf, *args)

    label = ROOT.Internal.RDF.GetDataSourceLabel(ROOT.RDF.AsRNode(localdf))
    if label == "TTreeDS":
        return TreeHeadNode(backend, npartitions, localdf, *args)
    elif label == "RNTupleDS":
        return RNTupleHeadNode(backend, npartitions, localdf, *args)
    elif label == "EmptyDS":
        return EmptySourceHeadNode(backend, npartitions, localdf, args[0])
    else:
        raise RuntimeError(
            (f"First argument {args[0]} of type {type(args[0])} is not "
             "recognised as a supported argument for distributed RDataFrame. "
             "Currently supported data sources are: TTree, RNTuple or an empty "
             "data source."))


class EmptySourceHeadNode(HeadNode):
    """
    The head node of a computation graph where the RDataFrame data source is
    empty and a number of sequential entries will be created at runtime. This
    head node is responsible for the following RDataFrame constructor::

        RDataFrame(ULong64_t numEntries)

    Attributes:
        nentries (int): The number of sequential entries the RDataFrame will create.

        npartitions (int): The number of partitions the dataset will be split in
            for distributed execution.
    """

    def __init__(self, backend: BaseBackend, npartitions: Optional[int], localdf: ROOT.RDataFrame, nentries: int):
        """
        Creates a new RDataFrame instance for the given arguments.

        Args:
            nentries (int): The number of entries this RDataFrame will process.

            npartitions (int): The number of partitions the dataset will be
                split in for distributed execution.
        """
        super().__init__(backend, npartitions, localdf)

        self.nentries = nentries

    def _build_ranges(self) -> List[Ranges.DataRange]:
        """Build the ranges for this dataset."""
        # Empty datasets cannot be processed distributedly
        if not self.nentries:
            raise RuntimeError(
                ("Cannot build a distributed RDataFrame with zero entries. "
                 "Distributed computation will fail. "))
        # TODO: This shouldn't be triggered if entries == 1. The current minimum
        # amount of partitions is 2. We need a robust reducer that smartly
        # becomes no-op if npartitions == 1 to avoid this.
        if self.npartitions > self.nentries:
            # Restrict 'npartitions' if it's greater than 'nentries'
            msg = ("Number of partitions {0} is greater than number of entries {1} "
                   "in the dataframe. Using {1} partition(s)".format(self.npartitions, self.nentries))
            warnings.warn(msg, UserWarning, stacklevel=2)
            self.npartitions = self.nentries
        return Ranges.get_balanced_ranges(self.nentries, self.npartitions, self.exec_id)

    def _generate_rdf_creator(self) -> Callable[[Ranges.DataRange], TaskObjects]:
        """
        Generates a function that is responsible for building an instance of
        RDataFrame on a distributed mapper for a given entry range. Specific for
        an empty data source.
        """

        nentries = self.nentries

        def build_rdf_from_range(current_range):
            """
            Builds an RDataFrame instance for a distributed mapper.
            """
            if current_range.exec_id not in _graph_cache._RDF_REGISTER:
                _graph_cache._RDF_REGISTER[current_range.exec_id] = ROOT.RDataFrame(nentries)

            ROOT.Internal.RDF.ChangeEmptyEntryRange(
                ROOT.RDF.AsRNode(_graph_cache._RDF_REGISTER[current_range.exec_id]),
                (current_range.start, current_range.end))

            return TaskObjects(_graph_cache._RDF_REGISTER[current_range.exec_id], None)

        return build_rdf_from_range

    def _handle_returned_values(self, values: TaskResult) -> Iterable:
        """
        Handle values returned after distributed execution. No extra checks are
        needed in the empty source case.
        """
        return values.mergeables


class TreeHeadNode(HeadNode):
    """
    The head node of a computation graph where the RDataFrame data source is
    a TTree or a TChain. This head node is responsible for the following
    RDataFrame constructors::

        RDataFrame(std::string_view treeName, std::string_view filenameglob, const ColumnNames_t &defaultBranches = {})
        RDataFrame(std::string_view treename, const std::vector<std::string> &fileglobs, const ColumnNames_t &defaultBranches = {})
        RDataFrame(std::string_view treeName, TDirectory *dirPtr, const ColumnNames_t &defaultBranches = {})
        RDataFrame(TTree &tree, const ColumnNames_t &defaultBranches = {})

    Attributes:
        npartitions (int): The number of partitions the dataset will be split in
            for distributed execution.

        tree (ROOT.TTree, ROOT.TChain): The dataset that will be processed. This
            is either stored as is if the user passed it as the first argument
            to the distributed RDataFrame constructor, or created from the
            arguments of the other constructor overloads.

        treename (str): The name of the dataset.

        inputfiles (list[str]): List of file names where the dataset is stored.

        defaultbranches (ROOT.std.vector[ROOT.std.string], None): Optional list
            of branches to be read from the tree passed by the user. Defaults to
            None.

        friendinfo (ROOT.Internal.TreeUtils.RFriendInfo, None): Optional
            information about friend trees of the dataset. Retrieved only if a
            TTree or TChain is passed to the constructor. Defaults to None.

    """

    def __init__(self, backend: BaseBackend, npartitions: Optional[int], localdf: ROOT.RDataFrame, *args):
        """
        Creates a new RDataFrame instance for the given arguments.

        Args:
            *args (iterable): Iterable with the arguments to the RDataFrame constructor.

            npartitions (int): The number of partitions the dataset will be
                split in for distributed execution.
        """
        super().__init__(backend, npartitions, localdf)

        self.defaultbranches = None
        # Information about friend trees, if they are present.
        self.friendinfo: Optional[ROOT.Internal.TreeUtils.RFriendInfo] = None

        # Retrieve the TTree/TChain that will be processed
        if isinstance(args[0], ROOT.TTree):
            # RDataFrame(tree, defaultBranches = {})
            tree = args[0]
            # TTreeIndex is not supported in distributed RDataFrame
            idx_found, friendname = ROOT.Internal.TreeUtils.TreeUsesIndexedFriends(
                tree)
            if idx_found:
                raise ValueError(
                    f"Friend tree '{friendname}' has a TTreeIndex. "
                    "This is not supported in distributed mode."
                )
            # Retrieve information about friend trees when user passes a TTree
            # or TChain object.
            fi = ROOT.Internal.TreeUtils.GetFriendInfo(args[0])
            self.friendinfo = fi if not fi.fFriendNames.empty() else None
            if len(args) == 2:
                self.defaultbranches = deepcopy(args[1])
        else:
            if isinstance(args[1], ROOT.TDirectory):
                # RDataFrame(treeName, dirPtr, defaultBranches = {})
                # We can assume both the argument TDirectory* and the TTree*
                # returned from TDirectory::Get are not nullptr since we already
                # did and early check of the user arguments in get_headnode
                tree = args[1].Get(args[0])
            elif isinstance(args[1], (str, ROOT.std.string_view)):
                # RDataFrame(treeName, filenameglob, defaultBranches = {})
                tree = ROOT.TChain(args[0])
                tree.Add(str(args[1]))
            elif isinstance(args[1], (list, ROOT.std.vector[ROOT.std.string])):
                # RDataFrame(treename, fileglobs, defaultBranches = {})
                tree = ROOT.TChain(args[0])
                for filename in args[1]:
                    tree.Add(str(filename))
            # In any of the three constructors considered in this branch, if
            # the user supplied three arguments then the third argument is a
            # list of default branches
            if len(args) == 3:
                self.defaultbranches = deepcopy(args[2])

        # maintreename: name of the tree or main name of the chain
        self.maintreename = str(tree.GetName())
        # subtreenames: names of all subtrees in the chain or full path to the tree in the file it belongs to
        self.subtreenames = [str(treename) for treename in ROOT.Internal.TreeUtils.GetTreeFullPaths(tree)]
        self.inputfiles = [str(filename) for filename in ROOT.Internal.TreeUtils.GetFileNamesFromTree(tree)]

    def _build_ranges(self) -> List[Ranges.DataRange]:
        """Build the ranges for this dataset."""
        logger.debug("Building ranges from dataset info:\n"
                     "main treename: %s\n"
                     "names of subtrees: %s\n"
                     "input files: %s\n", self.maintreename, self.subtreenames, self.inputfiles)

        if logger.isEnabledFor(logging.DEBUG):
            # Compute clusters and entries of the first tree in the dataset.
            # This will call once TFile::Open, but we pay this cost to get an estimate
            # on whether the number of requested partitions is reasonable.
            # Depending on the cluster setup, this may still be quite costly, so
            # we decide to pay the price only if the user explicitly requested
            # warning logging.
            clusters, entries = ROOT.Internal.TreeUtils.GetClustersAndEntries(self.subtreenames[0], self.inputfiles[0])
            # The file could contain an empty tree. In that case, the estimate will not be computed.
            if entries > 0:
                partitionsperfile = self.npartitions / len(self.inputfiles)
                if partitionsperfile > len(clusters):
                    logger.debug(
                        "The number of requested partitions could be higher than the maximum amount of "
                        "chunks the dataset can be split in. Some tasks could be doing no work. Consider "
                        "setting the 'npartitions' parameter of the RDataFrame constructor to a lower value.")

        return Ranges.get_percentage_ranges(self.subtreenames, self.inputfiles, self.npartitions, self.friendinfo, self.exec_id, None)

    def _generate_rdf_creator(self) -> Callable[[Ranges.DataRange], TaskObjects]:
        """
        Generates a function that is responsible for building an instance of
        RDataFrame on a distributed mapper for a given entry range. Specific for
        the TTree data source.
        """

        def attach_friend_info_if_present(current_range: Ranges.TreeRange,
                                          ds: ROOT.RDF.Experimental.RDatasetSpec) -> None:
            """
            Adds info about friend trees to the input chain. Also aligns the
            starting and ending entry of the friend chain cache to those of the
            main chain.
            """
            # Gather information about friend trees. Check that we got an
            # RFriendInfo struct and that it's not empty
            if (current_range.friendinfo is not None):
                # If the friend is a TChain, the zipped information looks like:
                # (name, alias), (file1.root, file2.root, ...), (subname1, subname2, ...)
                # If the friend is a TTree, the file list is made of
                # only one filename and the list of names of the sub trees
                # is empty, so the zipped information looks like:
                # (name, alias), (filename.root, ), ()
                zipped_friendinfo = zip(
                    current_range.friendinfo.fFriendNames,
                    current_range.friendinfo.fFriendFileNames,
                    current_range.friendinfo.fFriendChainSubNames
                )
                for (friend_name, friend_alias), friend_filenames, friend_chainsubnames in zipped_friendinfo:
                    friend_chainsubnames = friend_chainsubnames if len(friend_chainsubnames) > 0 else [friend_name]*len(friend_filenames)
                    ds.WithGlobalFriends(friend_chainsubnames, friend_filenames, friend_alias)

        def build_rdf_from_range(current_range: Ranges.TreeRangePerc) -> TaskObjects:
            """
            Builds an RDataFrame instance for a distributed mapper.

            The function creates a TChain from the information contained in the
            input range object. If the chain cannot be built, returns None.
            """

            clustered_range, entries_in_trees = Ranges.get_clustered_range_from_percs(current_range)

            if clustered_range is None:
                return TaskObjects(None, entries_in_trees)

            ds = ROOT.RDF.Experimental.RDatasetSpec()
            # add a sample with no name to represent the whole dataset
            ds.AddSample(("", clustered_range.treenames, clustered_range.filenames))
            ds.WithGlobalRange((clustered_range.globalstart, clustered_range.globalend))

            attach_friend_info_if_present(clustered_range, ds)

            if current_range.exec_id not in _graph_cache._RDF_REGISTER:
                # Fill the cache with the new RDataFrame
                _graph_cache._RDF_REGISTER[current_range.exec_id] = ROOT.RDataFrame(ds)
            else:
                # Update it to the range of entries for this task
                ROOT.Internal.RDF.ChangeSpec(
                    ROOT.RDF.AsRNode(_graph_cache._RDF_REGISTER[current_range.exec_id]),
                    ROOT.std.move(ds))

            return TaskObjects(_graph_cache._RDF_REGISTER[current_range.exec_id], entries_in_trees)

        return build_rdf_from_range

    def _handle_returned_values(self, values: TaskResult) -> Iterable:
        """
        Handle values returned after distributed execution. When the data source
        is a TTree, check that exactly the input files and all the entries in
        the dataset were processed during distributed execution.
        """
        if values.mergeables is None:
            raise RuntimeError("The distributed execution returned no values. "
                               "This can happen if all files in your dataset contain empty trees.")

        # User could have requested to read the same file multiple times indeed
        input_files_and_trees = [
            f"{filename}/{treename}" for filename, treename in zip(self.inputfiles, self.subtreenames)
        ]
        files_counts = Counter(input_files_and_trees)

        entries_in_trees = values.entries_in_trees
        # Keys should be exactly the same
        if files_counts.keys() != entries_in_trees.trees_with_entries.keys():
            raise RuntimeError("The specified input files and the files that were "
                                "actually processed are not the same:\n"
                                f"Input files: {list(files_counts.keys())}\n"
                                f"Processed files: {list(entries_in_trees.trees_with_entries.keys())}")

        # Multiply the entries of each tree by the number of times it was
        # requested by the user
        for fullpath in files_counts:
            entries_in_trees.trees_with_entries[fullpath] *= files_counts[fullpath]

        total_dataset_entries = sum(entries_in_trees.trees_with_entries.values())
        if entries_in_trees.processed_entries != total_dataset_entries:
            raise RuntimeError(f"The dataset has {total_dataset_entries} entries, "
                               f"but {entries_in_trees.processed_entries} were processed.")

        return values.mergeables
    
class RDatasetSpecHeadNode(HeadNode):
    """
    The head node of a computation graph where the RDataFrame data source is
    an RDatasetSpec object. This head node is responsible for the following
    RDataFrame constructors:
        RDataFrame(ROOT.RDF.Experimental.RDatasetSpec spec)

    Attributes:
        npartitions (int): The number of partitions the dataset will be split in
            for distributed execution.

        rdatasetspec (ROOT.RDF.Experimental.RDatasetSpec): rdataset spec object
            used to construct the RDataFrame

        subtreenames (list[str]): List of tree names in the dataset.
        
        inputfiles (list[str]): List of file names where the dataset is stored.

        friendinfo (ROOT.Internal.TreeUtils.RFriendInfo, None): Optional
            information about friend trees of the dataset. Retrieved from 
            RDatasetSpec.GetFriendInfo(). Defaults to None.
    """

    def __init__(self, backend: BaseBackend, npartitions: Optional[int], localdf: ROOT.RDataFrame, *args):
        """
        Creates a new RDataFrame instance for the given arguments.

        Args:
            *args (iterable): Iterable with the arguments to the RDataFrame constructor.

            npartitions (int): The number of partitions the dataset will be
                split in for distributed execution.
        """
        super().__init__(backend, npartitions, localdf)

        # Information about friend trees, if they are present.
        self.sampleMap: Dict[str, Ranges.SerializableRSample] = None 
        
        self.friendinfo: Optional[ROOT.Internal.TreeUtils.RFriendInfo] = None
                
            
        # Retrieve the RDatasetSpec that will be processed
        if isinstance(args[0], ROOT.RDF.Experimental.RDatasetSpec):
            # RDataFrame(rdatasetspec)
            fi = args[0].GetFriendInfo()
            self.friendinfo = fi if not fi.fFriendNames.empty() else None

        else:
            raise RuntimeError(
                f"First argument {args[0]} of type {type(args[0])} is not supported "
                "in RDatasetSpecHeaNode")
                
        if (args[0].GetEntryRangeBegin() != 0):
            warnings.warn("Setting global range is not supported in distributed parsing of RDatasetSpec, the set range will be ignored and all events will be processed.")
            
        # subtreenames: names of all subtrees in the chain or full path to the tree in the file it belongs to
        self.subtreenames = [str(treename) for treename in args[0].GetTreeNames()]
        self.inputfiles = [str(filename) for filename in args[0].GetFileNameGlobs()]
            
        # A map to map file id (filename/tree) to SerializableRSamples
        self.sampleMap = {}
        samples = ROOT.Internal.RDF.MoveOutSamples(args[0])

        for sample in samples:
            trees = sample.GetTreeNames()
            files = sample.GetFileNameGlobs()
            for file, tree in zip(files, trees): 
                sampleId = file + "/" + tree
                self.sampleMap.update({sampleId : Ranges.SerializableRSample(sample)})

    def _build_ranges(self) -> List[Ranges.DataRange]:
        """Build the ranges for this dataset."""
        logger.debug("Building ranges from dataset info:\n"
                     "names of subtrees: %s\n"
                     "input files: %s\n", self.subtreenames, self.inputfiles)

        if logger.isEnabledFor(logging.DEBUG):
            # Compute clusters and entries of the first tree in the dataset.
            # This will call once TFile::Open, but we pay this cost to get an estimate
            # on whether the number of requested partitions is reasonable.
            # Depending on the cluster setup, this may still be quite costly, so
            # we decide to pay the price only if the user explicitly requested
            # warning logging.
            clusters, entries = ROOT.Internal.TreeUtils.GetClustersAndEntries(
                self.subtreenames[0], self.inputfiles[0])
            # The file could contain an empty tree. In that case, the estimate will not be computed.
            if entries > 0:
                partitionsperfile = self.npartitions / len(self.inputfiles)
                if partitionsperfile > len(clusters):
                    logger.debug(
                        "The number of requested partitions could be higher than the maximum amount of "
                        "chunks the dataset can be split in. Some tasks could be doing no work. Consider "
                        "setting the 'npartitions' parameter of the RDataFrame constructor to a lower value.")
        
        return Ranges.get_percentage_ranges(self.subtreenames, self.inputfiles, self.npartitions, self.friendinfo, self.exec_id, self.sampleMap)

    def _generate_rdf_creator(self) -> Callable[[Ranges.DataRange], TaskObjects]:
        """
        Generates a function that is responsible for building an instance of
        RDataFrame on a distributed mapper for a given entry range.
        """

        def attach_friend_info_if_present(current_range: Ranges.TreeRange,
                                          ds: ROOT.RDF.Experimental.RDatasetSpec) -> None:
            """
            Adds info about friend trees to the input chain. Also aligns the
            starting and ending entry of the friend chain cache to those of the
            main chain.
            """
            # Gather information about friend trees. Check that we got an
            # RFriendInfo struct and that it's not empty
            if (current_range.friendinfo is not None):
                # If the friend is a TChain, the zipped information looks like:
                # (name, alias), (file1.root, file2.root, ...), (subname1, subname2, ...)
                # If the friend is a TTree, the file list is made of
                # only one filename and the list of names of the sub trees
                # is empty, so the zipped information looks like:
                # (name, alias), (filename.root, ), ()
                zipped_friendinfo = zip(
                    current_range.friendinfo.fFriendNames,
                    current_range.friendinfo.fFriendFileNames,
                    current_range.friendinfo.fFriendChainSubNames
                )
                for (friend_name, friend_alias), friend_filenames, friend_chainsubnames in zipped_friendinfo:
                    friend_chainsubnames = (
                        friend_chainsubnames if len(friend_chainsubnames) > 0
                        else [friend_name]*len(friend_filenames)
                    )
                    ds.WithGlobalFriends(
                        friend_chainsubnames, friend_filenames, friend_alias)
                    
        def build_rdf_from_range(current_range: Ranges.TreeRangePerc) -> TaskObjects:
            """
            Builds an RDataFrame instance for a distributed mapper.

            The function creates a TChain from the information contained in the
            input range object. If the chain cannot be built, returns None.
            """

            clustered_range, entries_in_trees = Ranges.get_clustered_range_from_percs(
                current_range)

            if clustered_range is None:
                return TaskObjects(None, entries_in_trees)
            
            ds = ROOT.RDF.Experimental.RDatasetSpec()
        
            filenames_cluster = clustered_range.filenames
            treenames_cluster = clustered_range.treenames

            # The goal is to correctly assign the metadata information to the event being
            # processed.
            # The distributed tasks are split based on entries in the processed files, we get the task ranges.
            # Files, however, also belong to various samples. We need to make sure the files are always
            # correctly matched with the samples they are coming from as this is essential for the correct metadata processing.
            # In order to do that, we need to find the common files (good_files) from the list of
            # files processed in a given task range (filenames_cluster) and the list of files belonging to the given sample
            # (mysample.GetFilenameGlobs). Additionally, we need a dictionary to correctly match files with their trees.

            # Matching files with trees 
            treesInFileMap = {}
            fileids_cluster = []
        
            for filename, treename in zip(filenames_cluster, treenames_cluster):
                fileid = filename + "/" + treename
                fileids_cluster.append(fileid)
                treesInFileMap.update({fileid : treename})

            # Making sure we don't double count samples  
            unique_samples = [] 
            samplenames = []
            
            for sample in clustered_range.samples:
                mysample = sample._sample
                samplename = mysample.GetSampleName()
                if samplename not in samplenames:
                    unique_samples.append(mysample)
                    samplenames.append(samplename)
            
            # Core: matching files with samples and adding samples to the spec that will be processed
            for mysample in unique_samples:
                sample_fileids = [
                    filenameglob + "/" + treename
                    for filenameglob, treename in zip(mysample.GetFileNameGlobs(), mysample.GetTreeNames())
                ]
                    
                good_files = list((Counter(filenames_cluster) & Counter(mysample.GetFileNameGlobs())).elements())
                good_fileids = list((Counter(fileids_cluster) & Counter(sample_fileids)).elements()) 

                good_trees =[
                    treesInFileMap.get(fileid) for fileid in good_fileids
                ]
                
                ds.AddSample((mysample.GetSampleName(), good_trees, good_files, mysample.GetMetaData()))
                
            ds.WithGlobalRange(
                (clustered_range.globalstart, clustered_range.globalend))
            
            attach_friend_info_if_present(clustered_range, ds)

            if current_range.exec_id not in _graph_cache._RDF_REGISTER:
                # Fill the cache with the new RDataFrame
                _graph_cache._RDF_REGISTER[current_range.exec_id] = ROOT.RDataFrame(ds) 
                
            else:
                # Update it to the range of entries for this task
                ROOT.Internal.RDF.ChangeSpec(
                    ROOT.RDF.AsRNode( _graph_cache._RDF_REGISTER[current_range.exec_id]),
                    ROOT.std.move(ds))

            return TaskObjects(_graph_cache._RDF_REGISTER[current_range.exec_id], entries_in_trees)

        return build_rdf_from_range

    def _handle_returned_values(self, values: TaskResult) -> Iterable:
        """
        Handle values returned after distributed execution. When the data source
        is a TTree, check that exactly the input files and all the entries in
        the dataset were processed during distributed execution.
        """
        if values.mergeables is None:
            raise RuntimeError("The distributed execution returned no values. "
                               "This can happen if all files in your dataset contain empty trees.")

        # User could have requested to read the same file multiple times indeed
        input_files_and_trees = [
            f"{filename}/{treename}" for filename, treename in zip(self.inputfiles, self.subtreenames)
        ]
        files_counts = Counter(input_files_and_trees)

        entries_in_trees = values.entries_in_trees
        # Keys should be exactly the same
        if files_counts.keys() != entries_in_trees.trees_with_entries.keys():
            raise RuntimeError("The specified input files and the files that were "
                               "actually processed are not the same:\n"
                               f"Input files: {list(files_counts.keys())}\n"
                               f"Processed files: {list(entries_in_trees.trees_with_entries.keys())}")

        # Multiply the entries of each tree by the number of times it was
        # requested by the user
        for fullpath in files_counts:
            entries_in_trees.trees_with_entries[fullpath] *= files_counts[fullpath]

        total_dataset_entries = sum(
            entries_in_trees.trees_with_entries.values())
        if entries_in_trees.processed_entries != total_dataset_entries:
            raise RuntimeError(f"The dataset has {total_dataset_entries} entries, "
                               f"but {entries_in_trees.processed_entries} were processed.")

        return values.mergeables
    
class RNTupleHeadNode(HeadNode):
    """
    The head node of a computation graph where the RDataFrame data source is
    an RNtuple.
    """

    def __init__(self, backend: BaseBackend, npartitions: Optional[int], localdf: ROOT.RDataFrame, *args):
        """
        Creates a new RDataFrame instance for the given arguments.

        Args:
            *args (iterable): Iterable with the arguments to the RDataFrame constructor.

            npartitions (int): The number of partitions the dataset will be
                split in for distributed execution.
        """
        super().__init__(backend, npartitions, localdf)

        self.mainntuplename = args[0]
        self.inputfiles = [args[1]] if isinstance(args[1], str) else args[1]
        # Keep this in accordance with the implementation of TTree for now
        self.subnames = [self.mainntuplename] * len(self.inputfiles)

    def _build_ranges(self) -> List[Ranges.DataRange]:
        # """Build the ranges for this dataset."""
        # For the moment, we explicitly pass only one "subname", since there is
        # only one name possible for the whole RNTuple
        # TODO: implement action cloning for RNTuple
        # Inject a sentinel flag in the execution identifier to specify this
        # RDataFrame instance is going to process RNTuple data and the computation
        # graph needs to be recreated at every task
        self.exec_id = _graph_cache.ExecutionIdentifier("RNTuple", self.exec_id.graph_uuid)
        return Ranges.get_percentage_ranges(self.subnames, self.inputfiles, self.npartitions, None, self.exec_id, None)

    def _generate_rdf_creator(self) -> Callable[[Ranges.DataRange], TaskObjects]:
        """
        Generates a function that is responsible for building an instance of
        RDataFrame on a distributed mapper for a given entry range. Specific for
        the RNtuple data source.
        """

        def build_rdf_from_range(current_range: Ranges.TreeRangePerc) -> TaskObjects:
            """
            Builds an RDataFrame instance for a distributed mapper.

            The function creates a TChain from the information contained in the
            input range object. If the chain cannot be built, returns None.
            """
            clustered_range, entries_in_rntuples = Ranges.get_clustered_range_from_percs(current_range)

            if clustered_range is None:
                return TaskObjects(None, entries_in_rntuples)

            return TaskObjects(
                ROOT.Internal.RDF.FromRNTuple(
                    clustered_range.treenames[0],  # Only one RNTuple name for the whole dataset
                    clustered_range.filenames,
                    (clustered_range.globalstart, clustered_range.globalend),
                ),
                
                entries_in_rntuples,
            )

        return build_rdf_from_range

    def _handle_returned_values(self, values: TaskResult) -> Iterable:
        """
        Handle values returned after distributed execution. When the data source
        is an RNtuple, check that exactly the input files and all the entries in
        the dataset were processed during distributed execution.
        """
        if values.mergeables is None:
            raise RuntimeError(
                "The distributed execution returned no values. "
                "This can happen if all files in your dataset contain empty trees."
            )

        # User could have requested to read the same file multiple times indeed
        input_files_and_trees = [
            f"{filename}/{ntuplename}" for filename, ntuplename in zip(self.inputfiles, self.subnames)
        ]
        files_counts = Counter(input_files_and_trees)

        entries_in_rntuples = values.entries_in_trees
        # Keys should be exactly the same
        if files_counts.keys() != entries_in_rntuples.trees_with_entries.keys():
            raise RuntimeError(
                "The specified input files and the files that were "
                "actually processed are not the same:\n"
                f"Input files: {list(files_counts.keys())}\n"
                f"Processed files: {list(entries_in_rntuples.trees_with_entries.keys())}"
            )

        # Multiply the entries of each tree by the number of times it was
        # requested by the user
        for fullpath in files_counts:
            entries_in_rntuples.trees_with_entries[fullpath] *= files_counts[fullpath]

        total_dataset_entries = sum(entries_in_rntuples.trees_with_entries.values())
        if entries_in_rntuples.processed_entries != total_dataset_entries:
            raise RuntimeError(f"The dataset has {total_dataset_entries} entries, "
                               f"but {entries_in_rntuples.processed_entries} were processed.")

        return values.mergeables        



## \file
## \ingroup tutorial_rcanvas
## The Higgs to two photons analysis from the ATLAS Open Data 2020 release, with RDataFrame.
##
## This tutorial is the Higgs to two photons analysis from the ATLAS Open Data release in 2020
## (http://opendata.atlas.cern/release/2020/documentation/). The data was taken with the ATLAS detector
## during 2016 at a center-of-mass energy of 13 TeV. Although the Higgs to two photons decay is very rare,
## the contribution of the Higgs can be seen as a narrow peak around 125 GeV because of the excellent
## reconstruction and identification efficiency of photons at the ATLAS experiment.
##
## The analysis is translated to a RDataFrame workflow processing 1.7 GB of simulated events and data.
##
## This macro is replica of tutorials/analysis/dataframe/df104_HiggsToTwoPhotons.py, but with usage of ROOT7 graphics
## Run macro with python3 -i df104.py command to get interactive canvas
##
## \macro_image (df104.png)
## \macro_code
##
## \date 2021-06-15
## \authors Stefan Wunsch (KIT, CERN) Sergey Linev (GSI)

import ROOT
import os
from ROOT.Experimental import (
    RCanvas,
    RText,
    RAttrText,
    RAttrFont,
    RAttrFill,
    RAttrLine,
    RLegend,
    RPadPos,
    RPadExtent,
    TObjectDrawable,
)

# Enable multi-threading
ROOT.ROOT.EnableImplicitMT()

# Create a ROOT dataframe for each dataset
path = "root://eospublic.cern.ch//eos/opendata/atlas/OutreachDatasets/2020-01-22"
df = {}
df["data"] = ROOT.RDataFrame(
    "mini", (os.path.join(path, "GamGam/Data/data_{}.GamGam.root".format(x)) for x in ("A", "B", "C", "D"))
)
df["ggH"] = ROOT.RDataFrame("mini", os.path.join(path, "GamGam/MC/mc_343981.ggH125_gamgam.GamGam.root"))
df["VBF"] = ROOT.RDataFrame("mini", os.path.join(path, "GamGam/MC/mc_345041.VBFH125_gamgam.GamGam.root"))
processes = list(df.keys())

# Apply scale factors and MC weight for simulated events and a weight of 1 for the data
for p in ["ggH", "VBF"]:
    df[p] = df[p].Define("weight", "scaleFactor_PHOTON * scaleFactor_PhotonTRIGGER * scaleFactor_PILEUP * mcWeight")
df["data"] = df["data"].Define("weight", "1.0")

# Select the events for the analysis
for p in processes:
    # Apply preselection cut on photon trigger
    df[p] = df[p].Filter("trigP")

    # Find two good muons with tight ID, pt > 25 GeV and not in the transition region between barrel and encap
    df[p] = (
        df[p]
        .Define(
            "goodphotons",
            "photon_isTightID && (photon_pt > 25000) && (abs(photon_eta) < 2.37) && ((abs(photon_eta) < 1.37) || (abs(photon_eta) > 1.52))",
        )
        .Filter("Sum(goodphotons) == 2")
    )

    # Take only isolated photons
    df[p] = (
        df[p]
        .Filter("Sum(photon_ptcone30[goodphotons] / photon_pt[goodphotons] < 0.065) == 2")
        .Filter("Sum(photon_etcone20[goodphotons] / photon_pt[goodphotons] < 0.065) == 2")
    )

# Compile a function to compute the invariant mass of the diphoton system
ROOT.gInterpreter.Declare(
    """
using Vec_t = const ROOT::VecOps::RVec<float>;
float ComputeInvariantMass(Vec_t& pt, Vec_t& eta, Vec_t& phi, Vec_t& e) {
    ROOT::Math::PtEtaPhiEVector p1(pt[0], eta[0], phi[0], e[0]);
    ROOT::Math::PtEtaPhiEVector p2(pt[1], eta[1], phi[1], e[1]);
    return (p1 + p2).mass() / 1000.0;
}
"""
)

# Define a new column with the invariant mass and perform final event selection
hists = {}
for p in processes:
    # Make four vectors and compute invariant mass
    df[p] = df[p].Define(
        "m_yy",
        "ComputeInvariantMass(photon_pt[goodphotons], photon_eta[goodphotons], photon_phi[goodphotons], photon_E[goodphotons])",
    )

    # Make additional kinematic cuts and select mass window
    df[p] = (
        df[p]
        .Filter("photon_pt[goodphotons][0] / 1000.0 / m_yy > 0.35")
        .Filter("photon_pt[goodphotons][1] / 1000.0 / m_yy > 0.25")
        .Filter("m_yy > 105 && m_yy < 160")
    )

    # Book histogram of the invariant mass with this selection
    hists[p] = df[p].Histo1D(
        ROOT.RDF.TH1DModel(p, "Diphoton invariant mass; m_{#gamma#gamma} [GeV];Events", 30, 105, 160), "m_yy", "weight"
    )

# Run the event loop

# RunGraphs allows to run the event loops of the separate RDataFrame graphs
# concurrently. This results in an improved usage of the available resources
# if each separate RDataFrame can not utilize all available resources, e.g.,
# because not enough data is available.
ROOT.RDF.RunGraphs([hists[s] for s in ["ggH", "VBF", "data"]])

ggh = hists["ggH"].GetValue()
vbf = hists["VBF"].GetValue()
data = hists["data"].GetValue()

# Create the plot

# Set styles - not yet available for v7
# ROOT.gROOT.SetStyle("ATLAS")

# Create canvas with pads for main plot and data/MC ratio
c = RCanvas.Create("df104_HiggsToTwoPhotons")

lower_pad = c.AddPad(RPadPos(0, 0.65), RPadExtent(1, 0.35))
upper_pad = c.AddPad(RPadPos(0, 0), RPadExtent(1, 0.65))

upper_frame = upper_pad.AddFrame()
upper_frame.margins.bottom = 0
upper_frame.margins.left = 0.14
upper_frame.margins.right = 0.05
upper_frame.x.labels.hide = True

lower_frame = lower_pad.AddFrame()
lower_frame.margins.top = 0
lower_frame.margins.left = 0.14
lower_frame.margins.right = 0.05
lower_frame.margins.bottom = 0.3

# Fit signal + background model to data
fit = ROOT.TF1("fit", "([0]+[1]*x+[2]*x^2+[3]*x^3)+[4]*exp(-0.5*((x-[5])/[6])^2)", 105, 160)
fit.FixParameter(5, 125.0)
fit.FixParameter(4, 119.1)
fit.FixParameter(6, 2.39)
fit.SetLineColor(2)
fit.SetLineStyle(ROOT.kSolid)
fit.SetLineWidth(2)
# do not draw fit function
data.Fit("fit", "0", "", 105, 160)

# Draw data
data.SetMarkerStyle(20)
data.SetMarkerSize(1.2)
data.SetLineWidth(2)
data.SetLineColor("black")
data.SetMinimum(1e-3)
data.SetMaximum(8e3)
data.GetYaxis().SetLabelSize(0.045)
data.GetYaxis().SetTitleSize(0.05)
data.GetYaxis().SetTitleOffset(1.4)
data.SetStats(False)
data.SetTitle("")

data_drawable = upper_pad.Add[TObjectDrawable]()
data_drawable.Set(data, "E")

# Draw fit
fit_drawable = upper_pad.Add[TObjectDrawable]()
fit_drawable.Set(fit, "SAME")

# Draw background
bkg = ROOT.TF1("bkg", "([0]+[1]*x+[2]*x^2+[3]*x^3)", 105, 160)
for i in range(4):
    bkg.SetParameter(i, fit.GetParameter(i))
bkg.SetLineColor(4)
bkg.SetLineStyle(ROOT.kDashed)
bkg.SetLineWidth(2)
bkg_drawable = upper_pad.Add[TObjectDrawable]()
bkg_drawable.Set(bkg, "SAME")

# Scale simulated events with luminosity * cross-section / sum of weights
# and merge to single Higgs signal
lumi = 10064.0
ggh.Scale(lumi * 0.102 / 55922617.6297)
vbf.Scale(lumi * 0.008518764 / 3441426.13711)
higgs = ggh.Clone()
higgs.Add(vbf)
higgs_drawable = upper_pad.Add[TObjectDrawable]()
higgs_drawable.Set(higgs, "HIST SAME")

# Draw ratio

ratiobkg = ROOT.TH1I("zero", "", 10, 105, 160)
ratiobkg.SetLineColor(4)
ratiobkg.SetLineStyle(ROOT.kDashed)
ratiobkg.SetLineWidth(2)
ratiobkg.SetMinimum(-125)
ratiobkg.SetMaximum(250)
ratiobkg.SetStats(False)
ratiobkg.GetXaxis().SetLabelSize(0.08)
ratiobkg.GetXaxis().SetTitleSize(0.12)
ratiobkg.GetXaxis().SetTitleOffset(1.0)
ratiobkg.GetYaxis().SetLabelSize(0.08)
ratiobkg.GetYaxis().SetTitleSize(0.09)
ratiobkg.GetYaxis().SetTitle("Data - Bkg.")
ratiobkg.GetYaxis().SetTitleOffset(0.7)
ratiobkg.GetYaxis().CenterTitle()
ratiobkg.GetYaxis().SetNdivisions(503, False)
ratiobkg.GetYaxis().ChangeLabel(-1, -1, 0)
ratiobkg.GetXaxis().SetTitle("m_{#gamma#gamma} [GeV]")
lower_pad.Add[TObjectDrawable]().Set(ratiobkg, "AXIS")

ratiosig = ROOT.TH1F("ratiosig", "ratiosig", 5500, 105, 160)
ratiosig.Eval(fit)
ratiosig.SetLineColor(2)
ratiosig.SetLineStyle(ROOT.kSolid)
ratiosig.SetLineWidth(2)
ratiosig.Add(bkg, -1)
lower_pad.Add[TObjectDrawable]().Set(ratiosig, "SAME")

ratiodata = data.Clone()
ratiodata.Add(bkg, -1)
for i in range(1, data.GetNbinsX()):
    ratiodata.SetBinError(i, data.GetBinError(i))

lower_pad.Add[TObjectDrawable]().Set(ratiodata, "E SAME")

# Add RLegend
legend = upper_pad.Draw[RLegend](RPadPos(0.01, 0.01), RPadExtent(0.3, 0.4))
legend.text.size = 0.05
legend.text.align = RAttrText.kRightCenter
legend.text.font = RAttrFont.kArial
legend.border.style = RAttrLine.kNone
legend.border.width = 0
legend.fill.style = RAttrFill.kHollow

legend.AddEntry(data_drawable, "Data", "lme")
legend.AddEntry(bkg_drawable, "Background", "l")
legend.AddEntry(fit_drawable, "Signal + Bkg.", "l")
legend.AddEntry(higgs_drawable, "Signal", "l")

# Add ATLAS labels
lbl1 = upper_pad.Draw[RText](RPadPos(0.05, 0.88), "ATLAS")
lbl1.onFrame = True
lbl1.text.font = RAttrFont.kArialBoldOblique
lbl1.text.size = 0.04
lbl1.text.align = RAttrText.kLeftBottom
lbl2 = upper_pad.Draw[RText](RPadPos(0.05 + 0.16, 0.88), "Open Data")
lbl2.onFrame = True
lbl2.text.font = RAttrFont.kArial
lbl2.text.size = 0.04
lbl2.text.align = RAttrText.kLeftBottom
lbl3 = upper_pad.Draw[RText](RPadPos(0.05, 0.82), "#sqrt{s} = 13 TeV, 10 fb^{-1}")
lbl3.onFrame = True
lbl3.text.font = RAttrFont.kArial
lbl3.text.size = 0.03
lbl3.text.align = RAttrText.kLeftBottom

# show canvas finally
c.SetSize(700, 780)
c.Show()

# Save plot in PNG file
if c.SaveAs("df104.png"):
    print("Saved figure to df104.png")



## \file
## \ingroup tutorial_roofit_main
## \notebook
## 'DATA AND CATEGORIES' RooFit tutorial macro #403
##
## Using weights in unbinned datasets
##
## \macro_image
## \macro_code
## \macro_output
##
## \date February 2018
## \authors Clemens Lange, Wouter Verkerke (C version)

import ROOT


# Create observable and unweighted dataset
# -------------------------------------------

# Declare observable
x = ROOT.RooRealVar("x", "x", -10, 10)
x.setBins(40)

# Construction a uniform pdf
p0 = ROOT.RooPolynomial("px", "px", x)

# Sample 1000 events from pdf
data = p0.generate(x, 1000)

# Calculate weight and make dataset weighted
# --------------------------------------------------

# Construct formula to calculate (fake) weight for events
wFunc = ROOT.RooFormulaVar("w", "event weight", "(x*x+10)", [x])

# Add column with variable w to previously generated dataset
data.addColumn(wFunc)

# Dataset d is now a dataset with two observable (x,w) with 1000 entries
data.Print()

# Create a new dataset wdata where w is interpreted as event weight rather than
# as observable
wdata = ROOT.RooDataSet(data.GetName(), data.GetTitle(), data.get(), Import=data, WeightVar="w")

# Dataset d is now a dataset with one observable (x) with 1000 entries and
# a sum of weights of ~430K
wdata.Print()

# Unbinned ML fit to weighted data
# ---------------------------------------------------------------

# Construction quadratic polynomial pdf for fitting
a0 = ROOT.RooRealVar("a0", "a0", 1)
a1 = ROOT.RooRealVar("a1", "a1", 0, -1, 1)
a2 = ROOT.RooRealVar("a2", "a2", 1, 0, 10)
p2 = ROOT.RooPolynomial("p2", "p2", x, [a0, a1, a2], 0)

# Fit quadratic polynomial to weighted data

# NOTE: A plain Maximum likelihood fit to weighted data does in general
#       NOT result in correct error estimates, individual
#       event weights represent Poisson statistics themselves.
#
# Fit with 'wrong' errors
r_ml_wgt = p2.fitTo(wdata, Save=True, PrintLevel=-1)

# A first order correction to estimated parameter errors in an
# (unbinned) ML fit can be obtained by calculating the
# covariance matrix as
#
#    V' = V C-1 V
#
# where V is the covariance matrix calculated from a fit
# to -logL = - sum [ w_i log f(x_i) ] and C is the covariance
# matrix calculated from -logL' = -sum [ w_i^2 log f(x_i) ]
# (i.e. the weights are applied squared)
#
# A fit in self mode can be performed as follows:

r_ml_wgt_corr = p2.fitTo(wdata, Save=True, SumW2Error=True, PrintLevel=-1)

# Plot weighted data and fit result
# ---------------------------------------------------------------

# Construct plot frame
frame = x.frame(Title="Unbinned ML fit, chi^2 fit to weighted data")

# Plot data using sum-of-weights-squared error rather than Poisson errors
wdata.plotOn(frame, DataError="SumW2")

# Overlay result of 2nd order polynomial fit to weighted data
p2.plotOn(frame)

# ML fit of pdf to equivalent unweighted dataset
# ---------------------------------------------------------------------

# Construct a pdf with the same shape as p0 after weighting
genPdf = ROOT.RooGenericPdf("genPdf", "x*x+10", [x])

# Sample a dataset with the same number of events as data
data2 = genPdf.generate(x, 1000)

# Sample a dataset with the same number of weights as data
data3 = genPdf.generate(x, 43000)

# Fit the 2nd order polynomial to both unweighted datasets and save the
# results for comparison
r_ml_unw10 = p2.fitTo(data2, Save=True, PrintLevel=-1)
r_ml_unw43 = p2.fitTo(data3, Save=True, PrintLevel=-1)

# Chis2 fit of pdf to binned weighted dataset
# ---------------------------------------------------------------------------

# Construct binned clone of unbinned weighted dataset
binnedData = wdata.binnedClone()
binnedData.Print("v")

# Perform chi2 fit to binned weighted dataset using sum-of-weights errors
#
# NB: Within the usual approximations of a chi2 fit, chi2 fit to weighted
# data using sum-of-weights-squared errors does give correct error
# estimates
chi2 = p2.createChi2(binnedData, ROOT.RooFit.DataError("SumW2"))
m = ROOT.RooMinimizer(chi2)
m.migrad()
m.hesse()

# Plot chi^2 fit result on frame as well
r_chi2_wgt = m.save()
p2.plotOn(frame, LineStyle="--", LineColor="r")

# Compare fit results of chi2, L fits to (un)weighted data
# ------------------------------------------------------------

# Note that ML fit on 1Kevt of weighted data is closer to result of ML fit on 43Kevt of unweighted data
# than to 1Kevt of unweighted data, the reference chi^2 fit with SumW2 error gives a result closer to
# that of an unbinned ML fit to 1Kevt of unweighted data.

print("==> ML Fit results on 1K unweighted events")
r_ml_unw10.Print()
print("==> ML Fit results on 43K unweighted events")
r_ml_unw43.Print()
print("==> ML Fit results on 1K weighted events with a summed weight of 43K")
r_ml_wgt.Print()
print("==> Corrected ML Fit results on 1K weighted events with a summed weight of 43K")
r_ml_wgt_corr.Print()
print("==> Chi2 Fit results on 1K weighted events with a summed weight of 43K")
r_chi2_wgt.Print()

c = ROOT.TCanvas("rf403_weightedevts", "rf403_weightedevts", 600, 600)
ROOT.gPad.SetLeftMargin(0.15)
frame.GetYaxis().SetTitleOffset(1.8)
frame.Draw()

c.SaveAs("rf403_weightedevts.png")



# \file
# \ingroup tutorial_roostats
# \notebook -js
# tutorial demonstrating and validates the RooJeffreysPrior class
#
# Jeffreys's prior is an 'objective prior' based on formal rules.
# It is calculated from the Fisher information matrix.
#
# Read more:
# http:#en.wikipedia.org/wiki/Jeffreys_prior
#
# The analytic form is not known for most PDFs, but it is for
# simple cases like the Poisson mean, Gaussian mean, Gaussian sigma.
#
# This class uses numerical tricks to calculate the Fisher Information Matrix
# efficiently.  In particular, it takes advantage of a property of the
# 'Asimov data' as described in
# Asymptotic formulae for likelihood-based tests of new physics
# Glen Cowan, Kyle Cranmer, Eilam Gross, Ofer Vitells
# http:#arxiv.org/abs/arXiv:1007.1727
#
# This Demo has four parts:
#  1. TestJeffreysPriorDemo -- validates Poisson mean case 1/sqrt(mu)
#  2. TestJeffreysGaussMean -- validates Gaussian mean case
#  3. TestJeffreysGaussSigma -- validates Gaussian sigma case 1/sigma
#  4. TestJeffreysGaussMeanAndSigma -- demonstrates 2-d example
#
# \macro_image
# \macro_output
# \macro_code
#
# \author Kyle Cranmer (C++ version), and P. P. (Python translation)

import ROOT


def rs302_JeffreysPriorDemo():

    w = ROOT.RooWorkspace("w")
    w.factory("Uniform::u(x[0,1])")
    w.factory("mu[100,1,200]")
    w.factory("ExtendPdf::p(u,mu)")

    asimov = w.pdf("p").generateBinned(w["x"], ExpectedData=True)

    res = w.pdf("p").fitTo(asimov, Save=True, SumW2Error=True)

    asimov.Print()
    res.Print()
    cov = res.covarianceMatrix()
    print("variance = ", (cov.Determinant()))
    print("stdev = ", cov.Determinant() ** 0.5)
    cov.Invert()
    print("jeffreys = ", cov.Determinant() ** 0.5)

    w.defineSet("poi", "mu")
    w.defineSet("obs", "x")

    pi = ROOT.RooJeffreysPrior("jeffreys", "jeffreys", w.pdf("p"), w.set("poi"), w.set("obs"))

    test = ROOT.RooGenericPdf("Expected", "Expected = 1/#sqrt#mu", "1./sqrt(mu)", w.set("poi"))

    c1 = ROOT.TCanvas()
    plot = w["mu"].frame()
    pi.plotOn(plot)
    test.plotOn(plot, LineColor="r", LineStyle=ROOT.kDashDotted)
    plot.Draw()

    legend = plot.BuildLegend()
    legend.DrawClone()

    c1.Update()
    c1.Draw()
    c1.SaveAs("rs302_JeffreysPriorDemo.1.png")


# _________________________________________________
def TestJeffreysGaussMean():

    w = ROOT.RooWorkspace("w")
    w.factory("Gaussian.g(x[0,-20,20],mu[0,-5.,5],sigma[1,0,10])")
    w.factory("n[10,.1,200]")
    w.factory("ExtendPdf.p(g,n)")
    w.var("sigma").setConstant()
    w.var("n").setConstant()

    asimov = w.pdf("p").generateBinned(w.var("x"), ExpectedData())

    resw.pdf("p").fitTo(asimov, Save(), SumW2Error(True))

    asimov.Print()
    res.Print()
    cov = res.covarianceMatrix()
    print("variance = ", (cov.Determinant()))
    print("stdev = ", sqrt(cov.Determinant()))
    cov.Invert()
    print("jeffreys = ", sqrt(cov.Determinant()))

    w.defineSet("poi", "mu")
    w.defineSet("obs", "x")

    pi = RooJeffreysPrior("jeffreys", "jeffreys", w.pdf("p"), w.set("poi"), w.set("obs"))

    temp = w.set("poi")
    pi.getParameters(temp).Print()

    #  return;
    test = RooGenericPdf("constant", "Expected = constant", "1", w.set("poi"))

    c2 = TCanvas()
    plot = w.var("mu").frame()
    pi.plotOn(plot)
    test.plotOn(plot, LineColor(kRed), LineStyle(kDashDotted))
    plot.Draw()

    legend = plot.BuildLegend()
    legend.DrawClone()
    c2.Update()
    c2.Draw()
    c2.SaveAs("rs302_JeffreysPriorDemo.2.png")


# _________________________________________________
def TestJeffreysGaussSigma():

    # this one is VERY sensitive
    # if the Gaussian is narrow ~ range(x)/nbins(x) then the peak isn't resolved
    #   and you get really bizarre shapes
    # if the Gaussian is too wide range(x) ~ sigma then PDF gets renormalized
    #   and the PDF falls off too fast at high sigma
    w = RooWorkspace("w")
    w.factory("Gaussian.g(x[0,-20,20],mu[0,-5,5],sigma[1,1,5])")
    w.factory("n[100,.1,2000]")
    w.factory("ExtendPdf.p(g,n)")
    #  w.var("sigma").setConstant()
    w.var("mu").setConstant()
    w.var("n").setConstant()
    w.var("x").setBins(301)

    asimov = w.pdf("p").generateBinned(w.var("x"), ExpectedData())

    resw.pdf("p").fitTo(asimov, Save(), SumW2Error(True))

    asimov.Print()
    res.Print()
    cov = res.covarianceMatrix()
    print("variance = ", (cov.Determinant()))
    print("stdev = ", sqrt(cov.Determinant()))
    cov.Invert()
    print("jeffreys = ", sqrt(cov.Determinant()))

    w.defineSet("poi", "sigma")
    w.defineSet("obs", "x")

    pi = RooJeffreysPrior("jeffreys", "jeffreys", w.pdf("p"), w.set("poi"), w.set("obs"))

    temp = w.set("poi")
    pi.getParameters(temp).Print()

    test = RooGenericPdf("test", "Expected = #sqrt2/#sigma", "sqrt(2.)/sigma", w.set("poi"))

    c3 = TCanvas()
    plot = w.var("sigma").frame()
    pi.plotOn(plot)
    test.plotOn(plot, LineColor(kRed), LineStyle(kDashDotted))
    plot.Draw()

    legend = plot.BuildLegend()
    legend.DrawClone()
    c3.Update()
    c3.Draw()
    c3.SaveAs("rs302_JeffreysPriorDemo.3.png")


# _________________________________________________
def TestJeffreysGaussMeanAndSigma():

    # this one is VERY sensitive
    # if the Gaussian is narrow ~ range(x)/nbins(x) then the peak isn't resolved
    #   and you get really bizarre shapes
    # if the Gaussian is too wide range(x) ~ sigma then PDF gets renormalized
    #   and the PDF falls off too fast at high sigma
    w = RooWorkspace("w")
    w.factory("Gaussian.g(x[0,-20,20],mu[0,-5,5],sigma[1,1.,5.])")
    w.factory("n[100,.1,2000]")
    w.factory("ExtendPdf.p(g,n)")

    w.var("n").setConstant()
    w.var("x").setBins(301)

    asimov = w.pdf("p").generateBinned(w.var("x"), ExpectedData())

    resw.pdf("p").fitTo(asimov, Save(), SumW2Error(True))

    asimov.Print()
    res.Print()
    cov = res.covarianceMatrix()
    print("variance = ", (cov.Determinant()))
    print("stdev = ", sqrt(cov.Determinant()))
    cov.Invert()
    print("jeffreys = ", sqrt(cov.Determinant()))

    w.defineSet("poi", "mu,sigma")
    w.defineSet("obs", "x")

    pi = RooJeffreysPrior("jeffreys", "jeffreys", w.pdf("p"), w.set("poi"), w.set("obs"))

    temp = w.set("poi")
    pi.getParameters(temp).Print()
    #  return;

    c4 = TCanvas()
    Jeff2d = pi.createHistogram(
        "2dJeffreys", w.var("mu"), Binning(10, -5.0, 5), YVar(w.var("sigma"), Binning(10, 1.0, 5.0))
    )
    Jeff2d.Draw("surf")
    c4.Update()
    c4.Draw()
    c4.SaveAs("rs302_JeffreysPriorDemo.4.png")


rs302_JeffreysPriorDemo()



## \file
## \ingroup tutorial_ml
## \notebook
##  TMVA Classification Example Using a Recurrent Neural Network
##
## This is an example of using a RNN in TMVA. We do classification using a toy time dependent data set
## that is generated when running this example macro
##
## \macro_code
##
## \author Harshal Shende


# TMVA Classification Example Using a Recurrent Neural Network

# This is an example of using a RNN in TMVA.
# We do the classification using a toy data set containing a time series of data sample ntimes
# and with dimension ndim that is generated when running the provided function `MakeTimeData (nevents, ntime, ndim)`


import ROOT

num_threads = 4  # use max 4 threads
# do enable MT running
if "imt" in ROOT.gROOT.GetConfigFeatures():
    ROOT.EnableImplicitMT(num_threads)
    # switch off MT in OpenBLAS to avoid conflict with tbb
    ROOT.gSystem.Setenv("OMP_NUM_THREADS", "1")
    print("Running with nthreads  = {}".format(ROOT.GetThreadPoolSize()))
else:
    print("Running in serial mode since ROOT does not support MT")


TMVA = ROOT.TMVA
TFile = ROOT.TFile

import os

TMVA.Tools.Instance()
TMVA.Config.Instance()


##  Helper function to generate the time data set
##  make some time data but not of fixed length.
##  use a poisson with mu = 5 and truncated at 10


def MakeTimeData(n, ntime, ndim):
    # ntime = 10;
    # ndim = 30; // number of dim/time

    fname = "time_data_t" + str(ntime) + "_d" + str(ndim) + ".root"
    v1 = []
    v2 = []

    for i in range(ntime):
        v1.append(ROOT.TH1D("h1_" + str(i), "h1", ndim, 0, 10))
        v2.append(ROOT.TH1D("h2_" + str(i), "h2", ndim, 0, 10))

    f1 = ROOT.TF1("f1", "gaus")
    f2 = ROOT.TF1("f2", "gaus")

    sgn = ROOT.TTree("sgn", "sgn")
    bkg = ROOT.TTree("bkg", "bkg")
    f = TFile(fname, "RECREATE")

    x1 = []
    x2 = []

    for i in range(ntime):
        x1.append(ROOT.std.vector["float"](ndim))
        x2.append(ROOT.std.vector["float"](ndim))

    for i in range(ntime):
        bkg.Branch("vars_time" + str(i), "std::vector<float>", x1[i])
        sgn.Branch("vars_time" + str(i), "std::vector<float>", x2[i])

    sgn.SetDirectory(f)
    bkg.SetDirectory(f)
    ROOT.gRandom.SetSeed(0)

    mean1 = ROOT.std.vector["double"](ntime)
    mean2 = ROOT.std.vector["double"](ntime)
    sigma1 = ROOT.std.vector["double"](ntime)
    sigma2 = ROOT.std.vector["double"](ntime)

    for j in range(ntime):
        mean1[j] = 5.0 + 0.2 * ROOT.TMath.Sin(ROOT.TMath.Pi() * j / float(ntime))
        mean2[j] = 5.0 + 0.2 * ROOT.TMath.Cos(ROOT.TMath.Pi() * j / float(ntime))
        sigma1[j] = 4 + 0.3 * ROOT.TMath.Sin(ROOT.TMath.Pi() * j / float(ntime))
        sigma2[j] = 4 + 0.3 * ROOT.TMath.Cos(ROOT.TMath.Pi() * j / float(ntime))

    for i in range(n):
        if i % 1000 == 0:
            print("Generating  event ... %d", i)

        for j in range(ntime):
            h1 = v1[j]
            h2 = v2[j]
            h1.Reset()
            h2.Reset()

            f1.SetParameters(1, mean1[j], sigma1[j])
            f2.SetParameters(1, mean2[j], sigma2[j])

            h1.FillRandom(f1, 1000)
            h2.FillRandom(f2, 1000)

            for k in range(ntime):
                # std::cout << j*10+k << "   ";
                x1[j][k] = h1.GetBinContent(k + 1) + ROOT.gRandom.Gaus(0, 10)
                x2[j][k] = h2.GetBinContent(k + 1) + ROOT.gRandom.Gaus(0, 10)

        sgn.Fill()
        bkg.Fill()

        if n == 1:
            c1 = ROOT.TCanvas()
            c1.Divide(ntime, 2)
            for j in range(ntime):
                c1.cd(j + 1)
                v1[j].Draw()
            for j in range(ntime):
                c1.cd(ntime + j + 1)
                v2[j].Draw()

            ROOT.gPad.Update()

    if n > 1:
        sgn.Write()
        bkg.Write()
        sgn.Print()
        bkg.Print()
        f.Close()


## macro for performing a classification using a Recurrent Neural Network
## @param use_type
##    use_type = 0    use Simple RNN network
##    use_type = 1    use LSTM network
##    use_type = 2    use GRU
##    use_type = 3    build 3 different networks with RNN, LSTM and GRU


use_type = 1
ninput = 30
ntime = 10
batchSize = 100
maxepochs = 10

nTotEvts = 2000  # total events to be generated for signal or background

useKeras = False

useTMVA_RNN = True
useTMVA_DNN = True
useTMVA_BDT = False

if ROOT.gSystem.GetFromPipe("root-config --has-tmva-pymva") == "yes":
    useKeras = True

if useKeras:
   try:
      pass
   except:
      ROOT.Warning("TMVA_RNN_Classification", "Skip using Keras since tensorflow cannot be imported")
      useKeras = False


rnn_types = ["RNN", "LSTM", "GRU"]
use_rnn_type = [1, 1, 1]

if 0 <= use_type < 3:
    use_rnn_type = [0, 0, 0]
    use_rnn_type[use_type] = 1

useGPU = True  # use GPU for TMVA if available

useGPU = "tmva-gpu" in ROOT.gROOT.GetConfigFeatures()
useTMVA_RNN = ("tmva-cpu" in ROOT.gROOT.GetConfigFeatures()) or useGPU

if useTMVA_RNN:
    ROOT.Warning(
        "TMVA_RNN_Classification",
        "TMVA is not build with GPU or CPU multi-thread support. Cannot use TMVA Deep Learning for RNN",
    )

archString = "GPU" if useGPU else "CPU"

writeOutputFile = True

rnn_type = "RNN"

if "tmva-pymva" in ROOT.gROOT.GetConfigFeatures():
    TMVA.PyMethodBase.PyInitialize()
else:
    useKeras = False



inputFileName = "time_data_t10_d30.root"

fileDoesNotExist = ROOT.gSystem.AccessPathName(inputFileName)

# if file does not exists create it
if fileDoesNotExist:
    MakeTimeData(nTotEvts, ntime, ninput)


inputFile = TFile.Open(inputFileName)
if inputFile is None:
    raise ROOT.Error("Error opening input file %s - exit", inputFileName.Data())


print("--- RNNClassification  : Using input file: {}".format(inputFile.GetName()))

# Create a ROOT output file where TMVA will store ntuples, histograms, etc.
outfileName = "data_RNN_" + archString + ".root"
outputFile = None


if writeOutputFile:
    outputFile = TFile.Open(outfileName, "RECREATE")


## Declare Factory

# Create the Factory class. Later you can choose the methods
# whose performance you'd like to investigate.

# The factory is the major TMVA object you have to interact with. Here is the list of parameters you need to
# pass

# - The first argument is the base of the name of all the output
# weightfiles in the directory weight/ that will be created with the
#     method parameters

# - The second argument is the output file for the training results
#
# - The third argument is a string option defining some general configuration for the TMVA session.
# For example all TMVA output can be suppressed by removing the "!" (not) in front of the "Silent" argument in
# the option string


# // Creating the factory object
factory = TMVA.Factory(
    "TMVAClassification",
    outputFile,
    V=False,
    Silent=False,
    Color=True,
    DrawProgressBar=True,
    Transformations=None,
    Correlations=False,
    AnalysisType="Classification",
    ModelPersistence=True,
)
dataloader = TMVA.DataLoader("dataset")

signalTree = inputFile.Get("sgn")
background = inputFile.Get("bkg")

nvar = ninput * ntime

## add variables - use new AddVariablesArray function
for i in range(ntime):
    dataloader.AddVariablesArray("vars_time" + str(i), ninput)


dataloader.AddSignalTree(signalTree, 1.0)
dataloader.AddBackgroundTree(background, 1.0)

# check given input
datainfo = dataloader.GetDataSetInfo()
vars = datainfo.GetListOfVariables()
print("number of variables is {}".format(vars.size()))


for v in vars:
    print(v)

nTrainSig = 0.8 * nTotEvts
nTrainBkg = 0.8 * nTotEvts

# Apply additional cuts on the signal and background samples (can be different)
mycuts = ""  # for example: TCut mycuts = "abs(var1)<0.5 && abs(var2-0.5)<1";
mycutb = ""

# build the string options for DataLoader::PrepareTrainingAndTestTree
dataloader.PrepareTrainingAndTestTree(
    mycuts,
    mycutb,
    nTrain_Signal=nTrainSig,
    nTrain_Background=nTrainBkg,
    SplitMode="Random",
    SplitSeed=100,
    NormMode="NumEvents",
    V=False,
    CalcCorrelations=False,
)

print("prepared DATA LOADER ")


## Book TMVA  recurrent models

# Book the different types of recurrent models in TMVA  (SimpleRNN, LSTM or GRU)


if useTMVA_RNN:
    for i in range(3):
        if not use_rnn_type[i]:
            continue

        rnn_type = rnn_types[i]

        ## Define RNN layer layout
        ##  it should be   LayerType (RNN or LSTM or GRU) |  number of units | number of inputs | time steps | remember output (typically no=0 | return full sequence
        rnnLayout = str(rnn_type) + "|10|" + str(ninput) + "|" + str(ntime) + "|0|1,RESHAPE|FLAT,DENSE|64|TANH,LINEAR"

        ## Defining Training strategies. Different training strings can be concatenate. Use however only one
        trainingString1 = "LearningRate=1e-3,Momentum=0.0,Repetitions=1,ConvergenceSteps=5,BatchSize=" + str(batchSize)
        trainingString1 += ",TestRepetitions=1,WeightDecay=1e-2,Regularization=None,MaxEpochs=" + str(maxepochs)
        trainingString1 += "Optimizer=ADAM,DropConfig=0.0+0.+0.+0."

        ## define the inputlayout string for RNN
        ## the input data should be organize as   following:
        ##/ input layout for RNN:    time x ndim
        ## add after RNN a reshape layer (needed top flatten the output) and a dense layer with 64 units and a last one
        ## Note the last layer is linear because  when using Crossentropy a Sigmoid is applied already
        ## Define the full RNN Noption string adding the final options for all network
        rnnName = "TMVA_" + str(rnn_type)
        factory.BookMethod(
            dataloader,
            TMVA.Types.kDL,
            rnnName,
            H=False,
            V=True,
            ErrorStrategy="CROSSENTROPY",
            VarTransform=None,
            WeightInitialization="XAVIERUNIFORM",
            ValidationSize=0.2,
            RandomSeed=1234,
            InputLayout=str(ntime) + "|" + str(ninput),
            Layout=rnnLayout,
            TrainingStrategy=trainingString1,
            Architecture=archString
        )


## Book TMVA  fully connected dense layer  models
if useTMVA_DNN:
    # Method DL with Dense Layer
    # Training strategies.
    trainingString1 = ROOT.TString(
        "LearningRate=1e-3,Momentum=0.0,Repetitions=1,"
        "ConvergenceSteps=10,BatchSize=256,TestRepetitions=1,"
        "WeightDecay=1e-4,Regularization=None,MaxEpochs=20"
        "DropConfig=0.0+0.+0.+0.,Optimizer=ADAM:"
    )  # + "|" + trainingString2
    # General Options.
    trainingString1.Append(archString)
    dnnName = "TMVA_DNN"
    factory.BookMethod(
        dataloader,
        TMVA.Types.kDL,
        dnnName,
        H=False,
        V=True,
        ErrorStrategy="CROSSENTROPY",
        VarTransform=None,
        WeightInitialization="XAVIER",
        RandomSeed=0,
        InputLayout="1|1|" + str(ntime * ninput),
        Layout="DENSE|64|TANH,DENSE|TANH|64,DENSE|TANH|64,LINEAR",
        TrainingStrategy=trainingString1
    )


## Book Keras recurrent models

# Book the different types of recurrent models in Keras  (SimpleRNN, LSTM or GRU)


if useKeras:
    for i in range(3):
        if use_rnn_type[i]:
            modelName = "model_" + rnn_types[i] + ".keras"
            trainedModelName = "trained_" + modelName
            print("Building recurrent keras model using a", rnn_types[i], "layer")
            # create python script which can be executed
            # create 2 conv2d layer + maxpool + dense
            # from keras.initializers import TruncatedNormal
            # from keras import initializations
            from tensorflow.keras.layers import (
                GRU,
                LSTM,
                Dense,
                Flatten,
                Reshape,
                SimpleRNN,
            )
            from tensorflow.keras.models import Sequential
            from tensorflow.keras.optimizers import Adam

            model = Sequential()
            model.add(Reshape((10, 30), input_shape=(10 * 30,)))
            # add recurrent neural network depending on type / Use option to return the full output
            if rnn_types[i] == "LSTM":
                model.add(LSTM(units=10, return_sequences=True))
            elif rnn_types[i] == "GRU":
                model.add(GRU(units=10, return_sequences=True))
            else:
                model.add(SimpleRNN(units=10, return_sequences=True))
            # m.AddLine("model.add(BatchNormalization())");
            model.add(Flatten())  # needed if returning the full time output sequence
            model.add(Dense(64, activation="tanh"))
            model.add(Dense(2, activation="sigmoid"))
            model.compile(loss="binary_crossentropy", optimizer=Adam(learning_rate=0.001), weighted_metrics=["accuracy"])
            model.save(modelName)
            model.summary()
            print("saved recurrent model", modelName)

            if not os.path.exists(modelName):
                useKeras = False
                print("Error creating Keras recurrent model file - Skip using Keras")
            else:
                # book PyKeras method only if Keras model could be created
                print("Booking Keras  model ", rnn_types[i])
                factory.BookMethod(
                    dataloader,
                    TMVA.Types.kPyKeras,
                    "PyKeras_" + rnn_types[i],
                    H=True,
                    V=False,
                    VarTransform=None,
                    FilenameModel=modelName,
                    FilenameTrainedModel="trained_" + modelName,
                    NumEpochs=maxepochs,
                    BatchSize=batchSize
                )


# use BDT in case not using Keras or TMVA DL
if not useKeras or not useTMVA_BDT:
    useTMVA_BDT = True


## Book TMVA BDT


if useTMVA_BDT:
    factory.BookMethod(
        dataloader,
        TMVA.Types.kBDT,
        "BDTG",
        H=True,
        V=False,
        NTrees=100,
        MinNodeSize="2.5%",
        BoostType="Grad",
        Shrinkage=0.10,
        UseBaggedBoost=True,
        BaggedSampleFraction=0.5,
        nCuts=20,
        MaxDepth=2,
    )


## Train all methods
factory.TrainAllMethods()

print("nthreads  = {}".format(ROOT.GetThreadPoolSize()))

# ---- Evaluate all MVAs using the set of test events
factory.TestAllMethods()

# ----- Evaluate and compare performance of all configured MVAs
factory.EvaluateAllMethods()

# check method

# plot ROC curve
c1 = factory.GetROCCurve(dataloader)
c1.Draw()

if outputFile:
    outputFile.Close()



### \file
### \ingroup tutorial_ml
### \notebook -nodraw
### Example of inference with SOFIE and RDataFrame, of a model trained with Keras.
### First, generate the input model by running `TMVA_Higgs_Classification.C`.
###
### This tutorial parses the input model and runs the inference using ROOT's JITing capability.
###
### \macro_code
### \macro_output
### \author Lorenzo Moneta

from os.path import exists

import ROOT

# check if the input file exists
modelFile = "HiggsModel.keras"
modelName = "HiggsModel"

if not exists(modelFile):
    raise FileNotFoundError("You need to run TMVA_Higgs_Classification.C to generate the Keras trained model")

# parse the input Keras model into RModel object
model = ROOT.TMVA.Experimental.SOFIE.PyKeras.Parse(modelFile)

# generating inference code
model.Generate()
model.OutputGenerated("Higgs_trained_model_generated.hxx")
model.PrintGenerated()

# compile using ROOT JIT trained model
print("compiling SOFIE model and functor....")
ROOT.gInterpreter.Declare('#include "Higgs_trained_model_generated.hxx"')
ROOT.gInterpreter.Declare('auto sofie_functor = TMVA::Experimental::SofieFunctor<7,TMVA_SOFIE_'+modelName+'::Session>(0,"Higgs_trained_model_generated.dat");')

# run inference over input data
inputFile = str(ROOT.gROOT.GetTutorialDir()) + "/machine_learning/data/Higgs_data.root"
df1 = ROOT.RDataFrame("sig_tree", inputFile)
h1 = df1.Define("DNN_Value", "sofie_functor(rdfslot_,m_jj, m_jjj, m_lv, m_jlv, m_bb, m_wbb, m_wwbb)").Histo1D(("h_sig", "", 100, 0, 1),"DNN_Value")

df2 = ROOT.RDataFrame("bkg_tree", inputFile)
h2 = df2.Define("DNN_Value", "sofie_functor(rdfslot_,m_jj, m_jjj, m_lv, m_jlv, m_bb, m_wbb, m_wwbb)").Histo1D(("h_bkg", "", 100, 0, 1),"DNN_Value")

# run over the input data once, combining both RDataFrame graphs.
ROOT.RDF.RunGraphs([h1, h2])

print("Number of signal entries",h1.GetEntries())
print("Number of background entries",h2.GetEntries())

h1.SetLineColor("kRed")
h2.SetLineColor("kBlue")

c1 = ROOT.TCanvas()
ROOT.gStyle.SetOptStat(0)

h2.DrawClone()
h1.DrawClone("SAME")
